# Copyright 2014 Intel
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

from ceilometer.ipmi.notifications import ironic as parser
from ceilometer.ipmi.platform import exception as ipmiexcept
from ceilometer.ipmi.platform import ipmi_sensor
from ceilometer.polling import plugin_base
from ceilometer import sample

LOG = log.getLogger(__name__)


class InvalidSensorData(ValueError):
    pass


class SensorPollster(plugin_base.PollsterBase):
    METRIC = None

    def setup_environment(self):
        super().setup_environment()
        self.ipmi = ipmi_sensor.IPMISensor()
        self.polling_failures = 0

        # Do not load this extension if no IPMI support
        if not self.ipmi.ipmi_support:
            raise plugin_base.ExtensionLoadError(
                "IPMITool not supported on host")

    @property
    def default_discovery(self):
        return 'local_node'

    @staticmethod
    def _get_sensor_types(data, sensor_type):
        # Ipmitool reports 'Pwr Consumption' as sensor type 'Current'.
        # Set sensor_type to 'Current' when polling 'Power' metrics.
        if sensor_type == 'Power':
            sensor_type = 'Current'

        try:
            return (sensor_type_data for _, sensor_type_data
                    in data[sensor_type].items())
        except KeyError:
            return []

    def get_samples(self, manager, cache, resources):
        # Only one resource for IPMI pollster
        try:
            stats = self.ipmi.read_sensor_any(self.METRIC)
        except ipmiexcept.IPMIException:
            self.polling_failures += 1
            LOG.warning(
                'Polling %(mtr)s sensor failed for %(cnt)s times!',
                {'mtr': self.METRIC, 'cnt': self.polling_failures})
            if 0 <= self.conf.ipmi.polling_retry < self.polling_failures:
                LOG.warning('Pollster for %s is disabled!', self.METRIC)
                raise plugin_base.PollsterPermanentError(resources)
            else:
                return

        self.polling_failures = 0

        sensor_type_data = self._get_sensor_types(stats, self.METRIC)

        for sensor_data in sensor_type_data:
            # Continue if sensor_data is not parseable.
            try:
                sensor_reading = sensor_data['Sensor Reading']
                sensor_id = sensor_data['Sensor ID']
            except KeyError:
                continue

            # Do not pick up power consumption metrics from 'Current' sensor
            if self.METRIC == 'Current' and 'Pwr Consumption' in sensor_id:
                continue

            if not parser.validate_reading(sensor_reading):
                continue

            try:
                volume, unit = parser.parse_reading(sensor_reading)
            except parser.InvalidSensorData:
                continue

            resource_id = '%(host)s-%(sensor-id)s' % {
                'host': self.conf.host,
                'sensor-id': parser.transform_id(sensor_id)
            }

            metadata = {
                'node': self.conf.host
            }

            extra_metadata = self.get_extra_sensor_metadata(sensor_data)
            if extra_metadata:
                metadata.update(extra_metadata)

            yield sample.Sample(
                name='hardware.ipmi.%s' % self.METRIC.lower(),
                type=sample.TYPE_GAUGE,
                unit=unit,
                volume=volume,
                user_id=None,
                project_id=None,
                resource_id=resource_id,
                resource_metadata=metadata)

    def get_extra_sensor_metadata(self, sensor_data):
        # override get_extra_sensor_metadata to add specific metrics for
        # each sensor
        return {}


class TemperatureSensorPollster(SensorPollster):
    METRIC = 'Temperature'


class CurrentSensorPollster(SensorPollster):
    METRIC = 'Current'


class FanSensorPollster(SensorPollster):
    METRIC = 'Fan'

    def get_extra_sensor_metadata(self, sensor_data):
        try:
            return {
                "maximum_rpm": sensor_data['Normal Maximum'],
            }
        except KeyError:
            # Maximum rpm might not be reported when usage
            # is reported as percent
            return {}


class VoltageSensorPollster(SensorPollster):
    METRIC = 'Voltage'


class PowerSensorPollster(SensorPollster):
    METRIC = 'Power'



#
# Copyright 2022 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ceilometer import cache_utils
from ceilometer import service as ceilometer_service
from oslo_cache.backends import dictionary
from oslo_cache import core as cache
from oslo_config import fixture as config_fixture
from oslotest import base


class CacheConfFixture(config_fixture.Config):
    def setUp(self):
        super().setUp()
        self.conf = ceilometer_service.\
            prepare_service(argv=[], config_files=[])
        cache.configure(self.conf)


class TestOsloCache(base.BaseTestCase):
    def setUp(self):
        super().setUp()

        conf = ceilometer_service.prepare_service(argv=[], config_files=[])

        dict_conf_fixture = CacheConfFixture(conf)
        self.useFixture(dict_conf_fixture)
        dict_conf_fixture.config(enabled=True, group='cache')
        dict_conf_fixture.config(expiration_time=600,
                                 backend='oslo_cache.dict',
                                 group='cache')
        self.dict_conf = dict_conf_fixture.conf

        # enable_retry_client is only supported by
        # 'dogpile.cache.pymemcache' backend which makes this
        # incorrect config
        faulty_conf_fixture = CacheConfFixture(conf)
        self.useFixture(faulty_conf_fixture)
        faulty_conf_fixture.config(enabled=True, group='cache')
        faulty_conf_fixture.config(expiration_time=600,
                                   backend='dogpile.cache.memcached',
                                   group='cache',
                                   enable_retry_client='true')
        self.faulty_conf = faulty_conf_fixture.conf

        no_cache_fixture = CacheConfFixture(conf)
        self.useFixture(no_cache_fixture)
        # no_cache_fixture.config()
        self.no_cache_conf = no_cache_fixture.conf

    def test_get_cache_region(self):
        self.assertIsNotNone(cache_utils.get_cache_region(self.dict_conf))

        # having invalid configurations will return None
        with self.assertLogs('ceilometer.cache_utils', level='ERROR') as logs:
            self.assertIsNone(
                cache_utils.get_cache_region(self.faulty_conf)
            )
            cache_configure_failed = logs.output
            self.assertIn(
                'ERROR:ceilometer.cache_utils:'
                'failed to configure oslo_cache: '
                'Retry client is only supported by '
                'the \'dogpile.cache.pymemcache\' backend.',
                cache_configure_failed)

    def test_get_client(self):
        dict_cache_client = cache_utils.get_client(self.dict_conf)
        self.assertIsNotNone(dict_cache_client)
        self.assertIsInstance(dict_cache_client.region.backend,
                              dictionary.DictCacheBackend)

        no_cache_config = cache_utils.get_client(self.no_cache_conf)
        self.assertIsNotNone(no_cache_config)
        self.assertIsInstance(dict_cache_client.region.backend,
                              dictionary.DictCacheBackend)

        # having invalid configurations will return None
        with self.assertLogs('ceilometer.cache_utils', level='ERROR') as logs:
            cache_client = cache_utils.get_client(self.faulty_conf)
            cache_configure_failed = logs.output
            self.assertIsNone(cache_client)
            self.assertIn(
                'ERROR:ceilometer.cache_utils:'
                'failed to configure oslo_cache: '
                'Retry client is only supported by '
                'the \'dogpile.cache.pymemcache\' backend.',
                cache_configure_failed)



#
# Copyright 2022 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Simple wrapper for oslo_cache."""
from keystoneauth1 import exceptions as ka_exceptions
from oslo_cache import core as cache
from oslo_cache import exception
from oslo_log import log

from ceilometer import keystone_client

# Default cache expiration period
CACHE_DURATION = 600

LOG = log.getLogger(__name__)


class CacheClient:
    def __init__(self, region, conf):
        self.region = region
        self.conf = conf

    def get(self, key):
        value = self.region.get(key)
        if value == cache.NO_VALUE:
            return None
        return value

    def set(self, key, value):
        return self.region.set(key, value)

    def delete(self, key):
        return self.region.delete(key)

    def resolve_uuid_from_cache(self, attr, uuid):
        resource_name = self.get(uuid)
        if resource_name:
            return resource_name
        else:
            # Retrieve project and user names from Keystone only
            # if ceilometer doesn't have a caching backend
            resource_name = self._resolve_uuid_from_keystone(attr, uuid)
            self.set(uuid, resource_name)
            return resource_name

    def _resolve_uuid_from_keystone(self, attr, uuid):
        try:
            return getattr(
                keystone_client.get_client(self.conf), attr
            ).get(uuid).name
        except AttributeError as e:
            LOG.warning("Found '%s' while resolving uuid %s to name", e, uuid)
        except ka_exceptions.NotFound as e:
            LOG.warning(e.message)


def get_client(conf):
    cache.configure(conf)
    if conf.cache.enabled:
        region = get_cache_region(conf)
        if region:
            return CacheClient(region, conf)
    else:
        # configure oslo_cache.dict backend if
        # no caching backend is configured
        region = get_dict_cache_region()
        return CacheClient(region, conf)


def get_dict_cache_region():
    region = cache.create_region()
    region.configure('oslo_cache.dict', expiration_time=CACHE_DURATION)
    return region


def get_cache_region(conf):
    # configure caching region using params from config
    try:
        region = cache.create_region()
        cache.configure_cache_region(conf, region)
        return region

    except exception.ConfigurationError as e:
        LOG.error("failed to configure oslo_cache: %s", str(e))



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara

# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""

import threading

from oslo_config import cfg
from oslo_utils import timeutils

OPTS = [
    cfg.StrOpt('rootwrap_config',
               default='/etc/ceilometer/rootwrap.conf',
               help='Path to the rootwrap configuration file to '
                    'use for running commands as root'),
]


def get_root_helper(conf):
    return 'sudo ceilometer-rootwrap %s' % conf.rootwrap_config


def spawn_thread(target, *args, **kwargs):
    t = threading.Thread(target=target, args=args, kwargs=kwargs)
    t.daemon = True
    t.start()
    return t


def isotime(at=None):
    """Current time as ISO string,

    :returns: Current time in ISO format
    """
    if not at:
        at = timeutils.utcnow()
    date_string = at.strftime("%Y-%m-%dT%H:%M:%S")
    tz = at.tzinfo.tzname(None) if at.tzinfo else 'UTC'
    date_string += ('Z' if tz == 'UTC' else tz)
    return date_string



# Copyright 2025 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1.access import service_catalog
from keystoneauth1.exceptions import catalog as catalog_exceptions
from oslo_config import cfg
from oslo_config import fixture as config_fixture

from ceilometer import keystone_client
from ceilometer import service as ceilo_service
from ceilometer.tests import base

CONF = cfg.CONF


class FakeSession:
    """A fake keystone auth session."""
    pass


class FakeAuthPlugin:
    """A fake keystone Auth Plugin."""
    pass


class TestKeystoneClient(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.CONF = ceilo_service.prepare_service([], [])
        # V3 catalog format
        catalog_data = [
            {
                'type': 'compute',
                'name': 'nova',
                'endpoints': [
                    {
                        'region_id': 'RegionOne',
                        'interface': 'public',
                        'url': 'http://nova.public/v2.1'
                    },
                    {
                        'region_id': 'RegionOne',
                        'interface': 'internal',
                        'url': 'http://nova.internal/v2.1'
                    },
                ]
            },
            {
                'type': 'identity',
                'name': 'keystone',
                'endpoints': [
                    {
                        'region_id': 'RegionOne',
                        'interface': 'public',
                        'url': 'http://keystone.public/v3'
                    },
                ]
            },
            {
                'type': 'network',
                'name': 'neutron',
                'endpoints': [
                    {
                        'region_id': 'RegionOne',
                        'interface': 'public',
                        'url': 'http://neutron.public'
                    }
                ]
            },
            {
                'type': 'radosgw',
                'name': 'radosgw',
                'endpoints': [
                    {
                        'region_id': 'RegionOne',
                        'interface': 'public',
                        'url': 'http://radosgw.public'
                    }
                ]
            },
            {
                'type': 'radosgw',
                'name': 'swift',
                'endpoints': [
                    {
                        'region_id': 'RegionOne',
                        'interface': 'public',
                        'url': 'http://swift.public'
                    }
                ]
            }
        ]
        self.fake_catalog = service_catalog.ServiceCatalogV3(catalog_data)

    @mock.patch(
        'keystoneauth1.loading.load_session_from_conf_options',
        return_value=FakeSession(), autospec=True)
    @mock.patch(
        'keystoneauth1.loading.load_auth_from_conf_options',
        return_value=FakeAuthPlugin(), autospec=True)
    def test_get_session(self, mock_load_auth, mock_load_session):

        session = keystone_client.get_session(self.CONF)

        mock_load_auth.assert_called_once_with(
            self.CONF, keystone_client.DEFAULT_GROUP)
        mock_load_session.assert_called_once_with(
            self.CONF, keystone_client.DEFAULT_GROUP,
            auth=mock_load_auth.return_value, session=None)
        self.assertIsInstance(session, FakeSession)

    @mock.patch(
        'keystoneauth1.loading.load_session_from_conf_options',
        return_value=FakeSession(), autospec=True)
    @mock.patch(
        'keystoneauth1.loading.load_auth_from_conf_options',
        return_value=FakeAuthPlugin(), autospec=True)
    def test_get_session_with_group(self, mock_load_auth, mock_load_session):

        session = keystone_client.get_session(
            self.CONF, group="some_other_group")

        mock_load_auth.assert_called_once_with(self.CONF, "some_other_group")
        mock_load_session.assert_called_once_with(
            self.CONF, "some_other_group",
            auth=mock_load_auth.return_value, session=None)
        self.assertIsInstance(session, FakeSession)

    @mock.patch(
        'keystoneauth1.loading.load_session_from_conf_options',
        return_value=FakeSession(), autospec=True)
    @mock.patch(
        'keystoneauth1.loading.load_auth_from_conf_options',
        return_value=FakeAuthPlugin(), autospec=True)
    def test_get_session_with_session(self, mock_load_auth, mock_load_session):
        fakeSession = FakeSession()

        session = keystone_client.get_session(self.CONF, fakeSession)

        mock_load_auth.assert_called_once_with(
            self.CONF, keystone_client.DEFAULT_GROUP)
        mock_load_session.assert_called_once_with(
            self.CONF, keystone_client.DEFAULT_GROUP,
            auth=mock_load_auth.return_value, session=fakeSession)
        self.assertIsInstance(session, FakeSession)

    @mock.patch(
        'keystoneauth1.loading.load_session_from_conf_options',
        return_value=FakeSession(), autospec=True)
    @mock.patch(
        'keystoneauth1.loading.load_auth_from_conf_options',
        return_value=FakeAuthPlugin(), autospec=True)
    def test_get_session_with_timeout(self, mock_load_auth, mock_load_session):

        session = keystone_client.get_session(self.CONF, timeout=100)

        mock_load_auth.assert_called_once_with(
            self.CONF, keystone_client.DEFAULT_GROUP)
        mock_load_session.assert_called_once_with(
            self.CONF, keystone_client.DEFAULT_GROUP,
            auth=mock_load_auth.return_value, session=None, timeout=100)
        self.assertIsInstance(session, FakeSession)

    @mock.patch('keystoneclient.v3.client.Client', autospec=True)
    @mock.patch('ceilometer.keystone_client.get_session', autospec=True)
    def test_get_client(self, mock_get_session, mock_ks_client):
        mock_session = FakeSession()
        mock_get_session.return_value = mock_session
        mock_client = mock.Mock()
        mock_ks_client.return_value = mock_client
        conf = self.useFixture(config_fixture.Config(self.CONF))
        conf.config(group=keystone_client.DEFAULT_GROUP, interface="internal")
        conf.config(
            group=keystone_client.DEFAULT_GROUP,
            region_name="expected_region")

        result = keystone_client.get_client(conf.conf)

        mock_get_session.assert_called_once_with(
            self.CONF,
            requests_session=None,
            group=keystone_client.DEFAULT_GROUP)
        mock_ks_client.assert_called_once_with(
            session=mock_get_session.return_value,
            trust_id=None,
            interface="internal",
            region_name="expected_region")
        self.assertEqual(result, mock_client)

    def test_get_service_catalog(self):
        mock_client = mock.Mock()
        mock_catalog = [
            {'name': 'keystone', 'type': 'identity'},
            {'name': 'nova', 'type': 'compute'}
        ]
        mock_access = mock.Mock()
        mock_access.service_catalog = mock_catalog
        mock_client.session.auth.get_access.return_value = mock_access

        result = keystone_client.get_service_catalog(mock_client)

        mock_client.session.auth.get_access.assert_called_once_with(
            mock_client.session)
        self.assertEqual(result, mock_catalog)

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        url = keystone_client.url_for(mock.Mock())

        self.assertEqual(url, "http://nova.public/v2.1")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_service_type(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        url = keystone_client.url_for(mock.Mock(), service_type="network")

        self.assertEqual(url, "http://neutron.public")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_service_name(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        url = keystone_client.url_for(mock.Mock(), service_name="neutron")

        self.assertEqual(url, "http://neutron.public")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_service_name_and_type_agree(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        url = keystone_client.url_for(
            mock.Mock(),
            service_type="compute",
            service_name="nova")

        self.assertEqual(url, "http://nova.public/v2.1")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_service_name_and_type_disagree(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        self.assertRaises(
            catalog_exceptions.EndpointNotFound,
            keystone_client.url_for,
            mock.Mock(),
            service_type="network",
            service_name="nova")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_non_existant_service(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        self.assertRaises(
            catalog_exceptions.EndpointNotFound,
            keystone_client.url_for,
            mock.Mock(),
            service_type="rating")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_interface(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        url = keystone_client.url_for(mock.Mock(), interface="public")

        self.assertEqual(url, "http://nova.public/v2.1")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_region_name(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        url = keystone_client.url_for(
            mock.Mock(),
            region_name="RegionOne")

        self.assertEqual(url, "http://nova.public/v2.1")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_region_name_non_existant(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        self.assertRaises(
            catalog_exceptions.EndpointNotFound,
            keystone_client.url_for,
            mock.Mock(),
            region_name="UnknownRegion")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_interface_internal(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        url = keystone_client.url_for(
            mock.Mock(),
            service_type="compute",
            interface="internal")

        self.assertEqual(url, "http://nova.internal/v2.1")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_interface_non_existant(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        self.assertRaises(
            catalog_exceptions.EndpointNotFound,
            keystone_client.url_for,
            mock.Mock(),
            service_type="network",
            interface="internal")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_service_type_and_interface(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        url = keystone_client.url_for(
            mock.Mock(),
            service_type="identity",
            interface="public")

        self.assertEqual(url, "http://keystone.public/v3")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_empty_catalog(self, mock_get_catalog):
        mock_get_catalog.return_value = service_catalog.ServiceCatalogV3([])

        self.assertRaises(
            catalog_exceptions.EndpointNotFound,
            keystone_client.url_for,
            mock.Mock(),
            service_type="compute")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_service_type_multiple_matches(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        url = keystone_client.url_for(
            mock.Mock(),
            service_type="radosgw")

        self.assertEqual(url, "http://radosgw.public")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_service_name_service_type_multi_match(
            self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        url = keystone_client.url_for(
            mock.Mock(),
            service_type="radosgw",
            service_name="swift")

        self.assertEqual(url, "http://swift.public")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_url_for_calls_get_service_catalog(self, mock_get_catalog):
        mock_catalog = mock.Mock()
        mock_catalog.url_for.return_value = "http://test.url"
        mock_get_catalog.return_value = mock_catalog
        mock_client = mock.Mock()

        result = keystone_client.url_for(
            mock_client,
            service_type="compute",
            service_name="nova",
            interface="public",
            region_name="RegionOne")

        mock_get_catalog.assert_called_once_with(mock_client)
        mock_catalog.url_for.assert_called_once_with(
            service_type="compute",
            service_name="nova",
            interface="public",
            region_name="RegionOne")
        self.assertEqual(result, "http://test.url")

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="compute",
            interface="public")

        self.assertEqual(urls, ("http://nova.public/v2.1",))

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_service_type(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="network",
            interface="public")

        self.assertEqual(urls, ("http://neutron.public",))

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_service_name(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_name="neutron",
            interface="public")

        self.assertEqual(urls, ("http://neutron.public",))

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_service_name_and_type_agree(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="compute",
            service_name="nova",
            interface="public")

        self.assertEqual(urls, ("http://nova.public/v2.1",))

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_service_name_and_type_disagree(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="network",
            service_name="nova",
            interface="public")

        self.assertEqual(urls, ())

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_non_existant_service(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="rating",
            interface="public")

        self.assertEqual(urls, ())

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_interface(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="compute",
            interface="public")

        self.assertEqual(urls, ("http://nova.public/v2.1",))

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_region_name(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="compute",
            interface="public",
            region_name="RegionOne")

        self.assertEqual(urls, ("http://nova.public/v2.1",))

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_region_name_non_existant(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="compute",
            interface="public",
            region_name="UnknownRegion")

        self.assertEqual(urls, ())

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_interface_internal(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="compute",
            interface="internal")

        self.assertEqual(urls, ("http://nova.internal/v2.1",))

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_interface_non_existant(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="network",
            interface="internal")

        self.assertEqual(urls, ())

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_service_type_and_interface(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="identity",
            interface="public")

        self.assertEqual(urls, ("http://keystone.public/v3",))

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_empty_catalog(self, mock_get_catalog):
        mock_get_catalog.return_value = service_catalog.ServiceCatalogV3([])

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="compute",
            interface="public")

        self.assertEqual(urls, ())

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_service_type_multiple_matches(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="radosgw")

        self.assertEqual(
            urls,
            ("http://radosgw.public", "http://swift.public"))

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_service_name_service_type_multi(self, mock_get_catalog):
        mock_get_catalog.return_value = self.fake_catalog

        urls = keystone_client.get_urls(
            mock.Mock(),
            service_type="radosgw",
            service_name="swift")

        self.assertEqual(
            urls, ("http://swift.public",))

    @mock.patch(
        'ceilometer.keystone_client.get_service_catalog', autospec=True)
    def test_get_urls_calls_get_service_catalog(self, mock_get_catalog):
        mock_catalog = mock.Mock()
        mock_catalog.get_urls.return_value = ["http://test.url"]
        mock_get_catalog.return_value = mock_catalog
        mock_client = mock.Mock()

        result = keystone_client.get_urls(
            mock_client,
            service_type="compute",
            service_name="nova",
            interface="public",
            region_name="RegionOne")

        mock_get_catalog.assert_called_once_with(mock_client)
        mock_catalog.get_urls.assert_called_once_with(
            service_type="compute",
            service_name="nova",
            interface="public",
            region_name="RegionOne")
        self.assertEqual(result, ["http://test.url"])

    def test_get_auth_token(self):
        mock_client = mock.Mock()
        mock_access = mock.Mock()
        mock_access.auth_token = 'auth_token'
        mock_client.session.auth.get_access.return_value = mock_access

        result = keystone_client.get_auth_token(mock_client)

        mock_client.session.auth.get_access.assert_called_once_with(
            mock_client.session)
        self.assertEqual(result, 'auth_token')



# Copyright 2012 eNovance <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import subprocess
import time

from oslo_utils import fileutils

from ceilometer.tests import base


class BinTestCase(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        content = ("[DEFAULT]\n"
                   "transport_url = fake://\n")
        content = content.encode('utf-8')
        self.tempfile = fileutils.write_to_tempfile(content=content,
                                                    prefix='ceilometer',
                                                    suffix='.conf')

    def tearDown(self):
        super().tearDown()
        os.remove(self.tempfile)

    def test_upgrade_run(self):
        subp = subprocess.Popen(['ceilometer-upgrade',
                                 '--skip-gnocchi-resource-types',
                                 "--config-file=%s" % self.tempfile])
        self.assertEqual(0, subp.wait())


class BinSendSampleTestCase(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        pipeline_cfg_file = self.path_get(
            'ceilometer/pipeline/data/pipeline.yaml')
        content = ("[DEFAULT]\n"
                   "transport_url = fake://\n"
                   "pipeline_cfg_file={}\n".format(pipeline_cfg_file))
        content = content.encode('utf-8')

        self.tempfile = fileutils.write_to_tempfile(content=content,
                                                    prefix='ceilometer',
                                                    suffix='.conf')

    def tearDown(self):
        super().tearDown()
        os.remove(self.tempfile)

    def test_send_counter_run(self):
        subp = subprocess.Popen(['ceilometer-send-sample',
                                 "--config-file=%s" % self.tempfile,
                                 "--sample-resource=someuuid",
                                 "--sample-name=mycounter"])
        self.assertEqual(0, subp.wait())


class BinCeilometerPollingServiceTestCase(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.tempfile = None
        self.subp = None

    def tearDown(self):
        if self.subp:
            try:
                self.subp.kill()
            except OSError:
                pass
        os.remove(self.tempfile)
        super().tearDown()

    def test_starting_with_duplication_namespaces(self):
        content = ("[DEFAULT]\n"
                   "transport_url = fake://\n")
        content = content.encode('utf-8')
        self.tempfile = fileutils.write_to_tempfile(content=content,
                                                    prefix='ceilometer',
                                                    suffix='.conf')
        self.subp = subprocess.Popen(['ceilometer-polling',
                                      "--config-file=%s" % self.tempfile,
                                      "--polling-namespaces",
                                      "compute",
                                      "compute"],
                                     stderr=subprocess.PIPE)
        expected = (b'Duplicated values: [\'compute\', \'compute\'] '
                    b'found in CLI options, auto de-duplicated')
        # NOTE(gordc): polling process won't quit so wait for a bit and check
        start = time.time()
        while time.time() - start < 5:
            output = self.subp.stderr.readline()
            if expected in output:
                break
        else:
            self.fail('Did not detect expected warning: %s' % expected)



# Copyright 2025 Catalyst Cloud Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from ceilometer.compute.pollsters import disk
from ceilometer.polling import manager
from ceilometer.tests.unit.compute.pollsters import base


class TestDiskPollsterBase(base.TestPollsterBase):

    TYPE = 'gauge'

    def setUp(self):
        super().setUp()
        self.instances = self._get_fake_instances()

    def _get_fake_instances(self, ephemeral=0):
        instances = []
        for i in [1, 2]:
            instance = mock.MagicMock()
            instance.name = f'instance-{i}'
            setattr(instance, 'OS-EXT-SRV-ATTR:instance_name',
                    instance.name)
            instance.id = i
            instance.flavor = {'name': 'm1.small', 'id': 2, 'vcpus': 1,
                               'ram': 512, 'disk': 20, 'ephemeral': ephemeral}
            instance.status = 'active'
            instances.append(instance)
        return instances

    def _check_get_samples(self,
                           factory,
                           name,
                           instances=None,
                           expected_count=2):
        pollster = factory(self.CONF)
        mgr = manager.AgentManager(0, self.CONF)
        samples = list(pollster.get_samples(mgr,
                                            {},
                                            instances or self.instances))
        self.assertGreater(len(samples), 0)
        self.assertEqual({name}, {s.name for s in samples},
                         (f"Only samples for meter {name} "
                          "should be published"))
        self.assertEqual(expected_count, len(samples))
        return samples


class TestDiskSizePollsters(TestDiskPollsterBase):

    TYPE = 'gauge'

    def test_ephemeral_disk_zero(self):
        samples = {
            sample.resource_id: sample
            for sample in self._check_get_samples(
                disk.EphemeralSizePollster,
                'disk.ephemeral.size',
                expected_count=len(self.instances))}
        for instance in self.instances:
            with self.subTest(instance.name):
                self.assertIn(instance.id, samples)
                sample = samples[instance.id]
                self.assertEqual(instance.flavor['ephemeral'],
                                 sample.volume)
                self.assertEqual(self.TYPE, sample.type)

    def test_ephemeral_disk_nonzero(self):
        instances = self._get_fake_instances(ephemeral=10)
        samples = {
            sample.resource_id: sample
            for sample in self._check_get_samples(
                disk.EphemeralSizePollster,
                'disk.ephemeral.size',
                instances=instances,
                expected_count=len(instances))}
        for instance in instances:
            with self.subTest(instance.name):
                self.assertIn(instance.id, samples)
                sample = samples[instance.id]
                self.assertEqual(instance.flavor['ephemeral'],
                                 sample.volume)
                self.assertEqual(self.TYPE, sample.type)

    def test_root_disk(self):
        samples = {
            sample.resource_id: sample
            for sample in self._check_get_samples(
                disk.RootSizePollster,
                'disk.root.size',
                expected_count=len(self.instances))}
        for instance in self.instances:
            with self.subTest(instance.name):
                self.assertIn(instance.id, samples)
                sample = samples[instance.id]
                self.assertEqual((instance.flavor['disk']
                                  - instance.flavor['ephemeral']),
                                 sample.volume)
                self.assertEqual(self.TYPE, sample.type)

    def test_root_disk_ephemeral_nonzero(self):
        instances = self._get_fake_instances(ephemeral=10)
        samples = {
            sample.resource_id: sample
            for sample in self._check_get_samples(
                disk.RootSizePollster,
                'disk.root.size',
                instances=instances,
                expected_count=len(instances))}
        for instance in instances:
            with self.subTest(instance.name):
                self.assertIn(instance.id, samples)
                sample = samples[instance.id]
                self.assertEqual((instance.flavor['disk']
                                  - instance.flavor['ephemeral']),
                                 sample.volume)
                self.assertEqual(self.TYPE, sample.type)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
from unittest import mock
import uuid

from oslo_utils import timeutils
from oslotest import base
from urllib import parse as urlparse

from ceilometer.event import models as event
from ceilometer.publisher import zaqar
from ceilometer import sample
from ceilometer import service


class TestZaqarPublisher(base.BaseTestCase):

    resource_id = str(uuid.uuid4())

    sample_data = [
        sample.Sample(
            name='alpha',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id=resource_id,
            timestamp=timeutils.utcnow().isoformat(),
            resource_metadata={'name': 'TestPublish'},
        ),
        sample.Sample(
            name='beta',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id=resource_id,
            timestamp=timeutils.utcnow().isoformat(),
            resource_metadata={'name': 'TestPublish'},
        ),
        sample.Sample(
            name='gamma',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id=resource_id,
            timestamp=datetime.datetime.now().isoformat(),
            resource_metadata={'name': 'TestPublish'},
        ),
    ]

    event_data = [event.Event(
        message_id=str(uuid.uuid4()), event_type='event_%d' % i,
        generated=timeutils.utcnow().isoformat(),
        traits=[], raw={'payload': {'some': 'aa'}}) for i in range(3)]

    def setUp(self):
        super().setUp()
        self.CONF = service.prepare_service([], [])

    def test_zaqar_publisher_config(self):
        """Test publisher config parameters."""
        parsed_url = urlparse.urlparse('zaqar://')
        self.assertRaises(ValueError, zaqar.ZaqarPublisher,
                          self.CONF, parsed_url)

        parsed_url = urlparse.urlparse('zaqar://?queue=foo&ttl=bar')
        self.assertRaises(ValueError, zaqar.ZaqarPublisher,
                          self.CONF, parsed_url)

        parsed_url = urlparse.urlparse('zaqar://?queue=foo&ttl=60')
        publisher = zaqar.ZaqarPublisher(self.CONF, parsed_url)
        self.assertEqual(60, publisher.ttl)

        parsed_url = urlparse.urlparse('zaqar://?queue=foo')
        publisher = zaqar.ZaqarPublisher(self.CONF, parsed_url)
        self.assertEqual(3600, publisher.ttl)
        self.assertEqual('foo', publisher.queue_name)

    @mock.patch('zaqarclient.queues.v2.queues.Queue')
    def test_zaqar_post_samples(self, mock_queue):
        """Test publisher post."""
        parsed_url = urlparse.urlparse('zaqar://?queue=foo')
        publisher = zaqar.ZaqarPublisher(self.CONF, parsed_url)
        mock_post = mock.Mock()
        mock_queue.return_value = mock_post

        publisher.publish_samples(self.sample_data)

        mock_queue.assert_called_once_with(mock.ANY, 'foo')
        self.assertEqual(
            3, len(mock_post.post.call_args_list[0][0][0]))
        self.assertEqual(
            mock_post.post.call_args_list[0][0][0][0]['body'],
            self.sample_data[0].as_dict())

    @mock.patch('zaqarclient.queues.v2.queues.Queue')
    def test_zaqar_post_events(self, mock_queue):
        """Test publisher post."""
        parsed_url = urlparse.urlparse('zaqar://?queue=foo')
        publisher = zaqar.ZaqarPublisher(self.CONF, parsed_url)
        mock_post = mock.Mock()
        mock_queue.return_value = mock_post

        publisher.publish_events(self.event_data)

        mock_queue.assert_called_once_with(mock.ANY, 'foo')
        self.assertEqual(
            3, len(mock_post.post.call_args_list[0][0][0]))
        self.assertEqual(
            mock_post.post.call_args_list[0][0][0][0]['body'],
            self.event_data[0].serialize())



#
# Copyright 2014 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from unittest import mock
import uuid

import fixtures
from gnocchiclient import exceptions as gnocchi_exc
from keystoneauth1 import exceptions as ka_exceptions
from oslo_config import fixture as config_fixture
from oslo_utils import fileutils
from oslo_utils import fixture as utils_fixture
from oslo_utils import netutils
from oslo_utils import timeutils
import requests
from stevedore import extension
import testscenarios

from ceilometer.event import models
from ceilometer.publisher import gnocchi
from ceilometer import sample
from ceilometer import service as ceilometer_service
from ceilometer.tests import base

load_tests = testscenarios.load_tests_apply_scenarios

INSTANCE_DELETE_START = models.Event(
    event_type='compute.instance.delete.start',
    traits=[models.Trait('state', 1, 'active'),
            models.Trait(
                'user_id', 1, '1e3ce043029547f1a61c1996d1a531a2'),
            models.Trait('service', 1, 'compute'),
            models.Trait('availability_zone', 1, 'zone1'),
            models.Trait('disk_gb', 2, 0),
            models.Trait('instance_type', 1, 'm1.tiny'),
            models.Trait('tenant_id', 1, '7c150a59fe714e6f9263774af9688f0e'),
            models.Trait('root_gb', 2, 0),
            models.Trait('ephemeral_gb', 2, 0),
            models.Trait('instance_type_id', 2, '2'),
            models.Trait('vcpus', 2, 1),
            models.Trait('memory_mb', 2, 512),
            models.Trait(
                'instance_id', 1, '9f9d01b9-4a58-4271-9e27-398b21ab20d1'),
            models.Trait('host', 1, 'vagrant-precise'),
            models.Trait(
                'request_id', 1, 'req-fb3c4546-a2e5-49b7-9fd2-a63bd658bc39'),
            models.Trait('project_id', 1, '7c150a59fe714e6f9263774af9688f0e'),
            models.Trait('launched_at', 4, '2012-05-08T20:23:47')],
    raw={},
    generated='2012-05-08T20:24:14.824743',
    message_id='a15b94ee-cb8e-4c71-9abe-14aa80055fb4',
)

INSTANCE_CREATE_END = models.Event(
    event_type='compute.instance.create.end',
    traits=[models.Trait('state', 1, 'active'),
            models.Trait(
                'user_id', 1, '1e3ce043029547f1a61c1996d1a531a2'),
            models.Trait('service', 1, 'compute'),
            models.Trait('availability_zone', 1, 'zone1'),
            models.Trait('disk_gb', 2, 0),
            models.Trait('instance_type', 1, 'm1.tiny'),
            models.Trait('tenant_id', 1, '7c150a59fe714e6f9263774af9688f0e'),
            models.Trait('root_gb', 2, 0),
            models.Trait('ephemeral_gb', 2, 0),
            models.Trait('instance_type_id', 2, '2'),
            models.Trait('vcpus', 2, 1),
            models.Trait('memory_mb', 2, 512),
            models.Trait(
                'instance_id', 1, '9f9d01b9-4a58-4271-9e27-398b21ab20d1'),
            models.Trait('host', 1, 'vagrant-precise'),
            models.Trait(
                'request_id', 1, 'req-fb3c4546-a2e5-49b7-9fd2-a63bd658bc39'),
            models.Trait('project_id', 1, '7c150a59fe714e6f9263774af9688f0e'),
            models.Trait('launched_at', 4, '2012-05-08T20:23:47')],
    raw={},
    generated='2012-05-08T20:24:14.824743',
    message_id='202f745e-4913-11e9-affe-9797342bd3a8',
)

IMAGE_DELETE_START = models.Event(
    event_type='image.delete',
    traits=[models.Trait('status', 1, 'deleted'),
            models.Trait('deleted_at', 1, '2016-11-04T04:25:56Z'),
            models.Trait('user_id', 1, 'e97ef33a20ed4843b520d223f3cc33d4'),
            models.Trait('name', 1, 'cirros'),
            models.Trait('service', 1, 'image.localhost'),
            models.Trait(
                'resource_id', 1, 'dc337359-de70-4044-8e2c-80573ba6e577'),
            models.Trait('created_at', 1, '2016-11-04T04:24:36Z'),
            models.Trait(
                'project_id', 1, 'e97ef33a20ed4843b520d223f3cc33d4'),
            models.Trait('size', 1, '13287936')],
    raw={},
    generated='2016-11-04T04:25:56.493820',
    message_id='7f5280f7-1d10-46a5-ba58-4d5508e49f99'
)


VOLUME_DELETE_END = models.Event(
    event_type='volume.delete.end',
    traits=[models.Trait('availability_zone', 1, 'nova'),
            models.Trait('created_at', 1, '2016-11-28T13:19:53+00:00'),
            models.Trait('display_name', 1, 'vol-001'),
            models.Trait(
                'host', 1, 'zhangguoqing-dev@lvmdriver-1#lvmdriver-1'),
            models.Trait(
                'project_id', 1, 'd53fcc7dc53c4662ad77822c36a21f00'),
            models.Trait('replication_status', 1, 'disabled'),
            models.Trait(
                'request_id', 1, 'req-f44df096-50d4-4211-95ea-64be6f5e4f60'),
            models.Trait(
                'resource_id', 1, '6cc6e7dd-d17d-460f-ae79-7e08a216ce96'),
            models.Trait(
                'service', 1, 'volume.zhangguoqing-dev@lvmdriver-1'),
            models.Trait('size', 1, '1'),
            models.Trait('status', 1, 'deleting'),
            models.Trait('tenant_id', 1, 'd53fcc7dc53c4662ad77822c36a21f00'),
            models.Trait('type', 1, 'af6271fa-13c4-44e6-9246-754ce9dc7df8'),
            models.Trait('user_id', 1, '819bbd28f5374506b8502521c89430b5')],
    raw={},
    generated='2016-11-28T13:42:15.484674',
    message_id='a15b94ee-cb8e-4c71-9abe-14aa80055fb4',
)

FLOATINGIP_DELETE_END = models.Event(
    event_type='floatingip.delete.end',
    traits=[models.Trait('service', 1, 'network.zhangguoqing-dev'),
            models.Trait(
                'project_id', 1, 'd53fcc7dc53c4662ad77822c36a21f00'),
            models.Trait(
                'request_id', 1, 'req-443ddb77-31f7-41fe-abbf-921107dd9f00'),
            models.Trait(
                'resource_id', 1, '705e2c08-08e8-45cb-8673-5c5be955569b'),
            models.Trait('tenant_id', 1, 'd53fcc7dc53c4662ad77822c36a21f00'),
            models.Trait('user_id', 1, '819bbd28f5374506b8502521c89430b5')],
    raw={},
    generated='2016-11-29T09:25:55.474710',
    message_id='a15b94ee-cb8e-4c71-9abe-14aa80055fb4'
)

VOLUME_TRANSFER_ACCEPT_END = models.Event(
    event_type='volume.transfer.accept.end',
    traits=[models.Trait('tenant_id', 1, '945e7d09220e4308abe4b3b734bf5fce>'),
            models.Trait('project_id', 1, '85bc015f7a2342348593077a927c4aaa'),
            models.Trait('user_id', 1, '945e7d09220e4308abe4b3b734bf5fce'),
            models.Trait('service', 1, 'volume.controller-0'),
            models.Trait(
                'request_id', 1, 'req-71dd1ae4-81ca-431a-b9fd-ac833eba889f'),
            models.Trait(
                'resource_id', 1, '156b8d3f-ad99-429b-b84c-3f263fb2a801'),
            models.Trait(
                'display_name', 1, 'test-vol'),
            models.Trait(
                'type', 1, 'req-71dd1ae4-81ca-431a-b9fd-ac833eba889f'),
            models.Trait('host', 1, 'hostgroup@tripleo_iscsi#tripleo_iscsi'),
            models.Trait('created_at', 4, '2020-08-28 12:51:52'),
            models.Trait('size', 2, 1)],
    raw={},
    generated='2020-08-28T12:52:22.930413',
    message_id='9fc4ceee-d980-4098-a685-2ad660838ac1'
)

SNAPSHOT_TRANSFER_ACCEPT_END = models.Event(
    event_type='snapshot.transfer.accept.end',
    traits=[models.Trait('tenant_id', 1, '945e7d09220e4308abe4b3b734bf5fce>'),
            models.Trait('project_id', 1, '85bc015f7a2342348593077a927c4aaa'),
            models.Trait('user_id', 1, '945e7d09220e4308abe4b3b734bf5fce'),
            models.Trait('service', 1, 'volume.controller-0'),
            models.Trait(
                'request_id', 1, 'req-71dd1ae4-81ca-431a-b9fd-ac833eba889f'),
            models.Trait(
                'resource_id', 1, '156b8d3f-ad99-429b-b84c-3f263fb2a801'),
            models.Trait(
                'display_name', 1, 'test-vol'),
            models.Trait(
                'type', 1, 'req-71dd1ae4-81ca-431a-b9fd-ac833eba889f'),
            models.Trait('host', 1, 'hostgroup@tripleo_iscsi#tripleo_iscsi'),
            models.Trait('created_at', 4, '2020-08-28 12:51:52'),
            models.Trait('size', 2, 1)],
    raw={},
    generated='2020-08-28T12:52:22.930413',
    message_id='9fc4ceee-d980-4098-a685-2ad660838ac1'
)


class PublisherTest(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        conf = ceilometer_service.prepare_service(argv=[], config_files=[])
        self.conf = self.useFixture(config_fixture.Config(conf))
        self.resource_id = str(uuid.uuid4())
        self.samples = [sample.Sample(
            name='disk.root.size',
            unit='GiB',
            type=sample.TYPE_GAUGE,
            volume=2,
            user_id='test_user',
            project_id='test_project',
            source='openstack',
            timestamp='2012-05-08 20:23:48.028195',
            resource_id=self.resource_id,
            resource_metadata={
                'host': 'foo',
                'image_ref': 'imageref!',
                'instance_flavor_id': 1234,
                'display_name': 'myinstance',
            }
        ),
            sample.Sample(
                name='disk.root.size',
                unit='GiB',
                type=sample.TYPE_GAUGE,
                volume=2,
                user_id='test_user',
                project_id='test_project',
                source='openstack',
                timestamp='2014-05-08 20:23:48.028195',
                resource_id=self.resource_id,
                resource_metadata={
                    'host': 'foo',
                    'image_ref': 'imageref!',
                    'instance_flavor_id': 1234,
                    'display_name': 'myinstance',
                },
        ),
        ]

        ks_client = mock.Mock(auth_token='fake_token')
        ks_client.projects.find.return_value = mock.Mock(
            name='gnocchi', id='a2d42c23-d518-46b6-96ab-3fba2e146859')
        self.useFixture(fixtures.MockPatch(
            'ceilometer.keystone_client.get_client',
            return_value=ks_client))
        self.useFixture(fixtures.MockPatch(
            'gnocchiclient.v1.client.Client',
            return_value=mock.Mock()))
        self.useFixture(fixtures.MockPatch(
            'ceilometer.keystone_client.get_session',
            return_value=mock.Mock()))
        self.ks_client = ks_client

    def test_config_load(self):
        url = netutils.urlsplit("gnocchi://")
        d = gnocchi.GnocchiPublisher(self.conf.conf, url)
        names = [rd.cfg['resource_type'] for rd in d.resources_definition]
        self.assertIn('instance', names)
        self.assertIn('volume', names)

    def test_match(self):
        resource = {
            'metrics':
                ['image', 'image.size', 'image.download', 'image.serve'],
            'attributes':
                {'container_format': 'resource_metadata.container_format',
                 'disk_format': 'resource_metadata.disk_format',
                 'name': 'resource_metadata.name'},
            'event_delete': 'image.delete',
            'event_attributes': {'id': 'resource_id'},
            'resource_type': 'image'}
        plugin_manager = extension.ExtensionManager(
            namespace='ceilometer.event.trait.trait_plugin')
        rd = gnocchi.ResourcesDefinition(
            resource, "high", "low", plugin_manager)
        operation = rd.event_match("image.delete")
        self.assertEqual('delete', operation)

    def test_metric_match(self):
        pub = gnocchi.GnocchiPublisher(self.conf.conf,
                                       netutils.urlsplit("gnocchi://"))
        self.assertIn('image.size', pub.metric_map['image.size'].metrics)

    @mock.patch('ceilometer.publisher.gnocchi.LOG')
    def test_broken_config_load(self, mylog):
        contents = [("---\n"
                     "resources:\n"
                     "  - resource_type: foobar\n"),
                    ("---\n"
                     "resources:\n"
                     "  - resource_type: 0\n"),
                    ("---\n"
                     "resources:\n"
                     "  - sample_types: ['foo', 'bar']\n"),
                    ("---\n"
                     "resources:\n"
                     "  - sample_types: foobar\n"
                     "  - resource_type: foobar\n"),
                    ]

        for content in contents:
            content = content.encode('utf-8')

            temp = fileutils.write_to_tempfile(content=content,
                                               prefix='gnocchi_resources',
                                               suffix='.yaml')
            self.addCleanup(os.remove, temp)
            url = netutils.urlsplit(
                "gnocchi://?resources_definition_file=" + temp)
            d = gnocchi.GnocchiPublisher(self.conf.conf, url)
            self.assertTrue(mylog.error.called)
            self.assertEqual(0, len(d.resources_definition))

    @mock.patch('ceilometer.publisher.gnocchi.GnocchiPublisher'
                '._if_not_cached', mock.Mock())
    @mock.patch('ceilometer.publisher.gnocchi.GnocchiPublisher'
                '.batch_measures')
    def _do_test_activity_filter(self,
                                 expected_measures,
                                 fake_batch,
                                 params=None):
        url_str = "gnocchi://"
        if params:
            url_str += "?" + "&".join(f"{k}={v}" for k, v in params.items())
        url = netutils.urlsplit(url_str)
        d = gnocchi.GnocchiPublisher(self.conf.conf, url)
        d._already_checked_archive_policies = True
        d.publish_samples(self.samples)
        self.assertEqual(1, len(fake_batch.mock_calls))
        measures = fake_batch.mock_calls[0][1][0]
        self.assertEqual(
            expected_measures,
            sum(len(m["measures"]) for rid in measures
                for m in measures[rid].values()))

    def test_activity_filter_match_project_id(self):
        self.samples[0].project_id = (
            'a2d42c23-d518-46b6-96ab-3fba2e146859')
        self._do_test_activity_filter(1)

    @mock.patch('ceilometer.publisher.gnocchi.LOG')
    def test_activity_gnocchi_project_not_found(self, logger):
        self.ks_client.projects.find.side_effect = ka_exceptions.NotFound
        self._do_test_activity_filter(2)
        log_called_args = logger.warning.call_args_list
        self.assertEqual(
            'Filtered project [service] not found in keystone, ignoring the '
            'filter_project option',
            log_called_args[0][0][0] % log_called_args[0][0][1])

    @mock.patch('ceilometer.publisher.gnocchi.LOG')
    def test_activity_gnocchi_project_filter_disabled(self, logger):
        self.samples[0].project_id = (
            'a2d42c23-d518-46b6-96ab-3fba2e146859')
        self._do_test_activity_filter(
            2,
            params={"enable_filter_project": "false"})
        self.assertIn(
            mock.call("Filtering Gnocchi project is disabled, "
                      "ignoring filter_project option"),
            logger.debug.call_args_list,
        )

    @mock.patch('ceilometer.publisher.gnocchi.GnocchiPublisher'
                '._get_gnocchi_client')
    def test_get_gnocchi_client(self, gnocchi_cli):
        url = netutils.urlsplit("gnocchi://")
        gnocchi_cli.side_effect = ka_exceptions.DiscoveryFailure
        cfg = self.conf.conf
        publisher = gnocchi.GnocchiPublisher
        self.assertRaises(ka_exceptions.DiscoveryFailure, publisher, cfg, url)

    def test_activity_filter_match_swift_event(self):
        self.samples[0].name = 'storage.objects.outgoing.bytes'
        self.samples[0].resource_id = 'a2d42c23-d518-46b6-96ab-3fba2e146859'
        self._do_test_activity_filter(1)

    def test_activity_filter_nomatch(self):
        self._do_test_activity_filter(2)

    @mock.patch('ceilometer.publisher.gnocchi.GnocchiPublisher'
                '.batch_measures')
    def test_unhandled_meter(self, fake_batch):
        samples = [sample.Sample(
            name='unknown.meter',
            unit='GiB',
            type=sample.TYPE_GAUGE,
            volume=2,
            user_id='test_user',
            project_id='test_project',
            source='openstack',
            timestamp='2014-05-08 20:23:48.028195',
            resource_id='randomid',
            resource_metadata={}
        )]
        url = netutils.urlsplit("gnocchi://")
        d = gnocchi.GnocchiPublisher(self.conf.conf, url)
        d._already_checked_archive_policies = True
        d.publish_samples(samples)
        self.assertEqual(0, len(fake_batch.call_args[0][1]))

    @mock.patch('ceilometer.publisher.gnocchi.GnocchiPublisher'
                '.batch_measures')
    def test_unhandled_meter_with_no_resource_id(self, fake_batch):
        samples = [
            sample.Sample(
                name='unknown.meter',
                unit='GiB',
                type=sample.TYPE_GAUGE,
                volume=2,
                user_id='test_user',
                project_id='test_project',
                source='openstack',
                timestamp='2014-05-08 20:23:48.028195',
                resource_id=None,
                resource_metadata={}),
            sample.Sample(
                name='unknown.meter',
                unit='GiB',
                type=sample.TYPE_GAUGE,
                volume=2,
                user_id='test_user',
                project_id='test_project',
                source='openstack',
                timestamp='2014-05-08 20:23:48.028195',
                resource_id="Some-other-resource-id",
                resource_metadata={})
        ]
        url = netutils.urlsplit("gnocchi://")
        d = gnocchi.GnocchiPublisher(self.conf.conf, url)
        d._already_checked_archive_policies = True
        d.publish_samples(samples)
        self.assertEqual(0, len(fake_batch.call_args[0][1]))

    @mock.patch('ceilometer.publisher.gnocchi.LOG')
    @mock.patch('gnocchiclient.v1.client.Client')
    def test__set_update_attributes_non_existent_resource(self,
                                                          fakeclient_cls,
                                                          logger):
        url = netutils.urlsplit("gnocchi://")
        self.publisher = gnocchi.GnocchiPublisher(self.conf.conf, url)

        fakeclient = fakeclient_cls.return_value
        fakeclient.resource.update.side_effect = [
            gnocchi_exc.ResourceNotFound(404)]

        non_existent_resource = {
            'type': 'volume',
            'id': self.resource_id,
        }

        self.publisher._set_update_attributes(non_existent_resource)

        logger.debug.assert_called_with(
            "Update event received on unexisting resource (%s), ignore it.",
            self.resource_id)

    def test_stable_resource_attributes_hash(self):
        url = netutils.urlsplit("gnocchi://")
        publisher = gnocchi.GnocchiPublisher(self.conf.conf, url)
        attributes = {'hello': 'world', 'foo': 'bar'}
        hash = publisher._hash_resource(attributes)
        self.assertEqual(hash, publisher._hash_resource(attributes))


class MockResponse(mock.NonCallableMock):
    def __init__(self, code):
        text = {500: 'Internal Server Error',
                404: 'Not Found',
                204: 'Created',
                409: 'Conflict',
                }.get(code)
        super().__init__(spec=requests.Response,
                         status_code=code,
                         text=text)


class PublisherWorkflowTest(base.BaseTestCase,
                            testscenarios.TestWithScenarios):

    sample_scenarios = [
        ('cpu', dict(
            sample=sample.Sample(
                resource_id=str(uuid.uuid4()) + "_foobar",
                name='cpu',
                unit='ns',
                type=sample.TYPE_CUMULATIVE,
                volume=500,
                user_id='test_user',
                project_id='test_project',
                source='openstack',
                timestamp='2012-05-08 20:23:48.028195',
                resource_metadata={
                    'host': 'foo',
                    'image_ref': 'imageref!',
                    'instance_flavor_id': 1234,
                    'display_name': 'myinstance',
                },
            ),
            metric_attributes={
                "archive_policy_name": "ceilometer-low-rate",
                "unit": "ns",
                "measures": [{
                    'timestamp': '2012-05-08 20:23:48.028195',
                    'value': 500
                }]
            },
            postable_attributes={
                'user_id': 'test_user',
                'project_id': 'test_project',
            },
            patchable_attributes={
                'host': 'foo',
                'image_ref': 'imageref!',
                'flavor_id': 1234,
                'display_name': 'myinstance',
            },
            resource_type='instance')),
        ('disk.root.size', dict(
            sample=sample.Sample(
                resource_id=str(uuid.uuid4()) + "_foobar",
                name='disk.root.size',
                unit='GiB',
                type=sample.TYPE_GAUGE,
                volume=2,
                user_id='test_user',
                project_id='test_project',
                source='openstack',
                timestamp='2012-05-08 20:23:48.028195',
                resource_metadata={
                    'host': 'foo',
                    'image_ref': 'imageref!',
                    'instance_flavor_id': 1234,
                    'display_name': 'myinstance',
                },
            ),
            metric_attributes={
                "archive_policy_name": "ceilometer-low",
                "unit": "GiB",
                "measures": [{
                    'timestamp': '2012-05-08 20:23:48.028195',
                    'value': 2
                }]
            },
            postable_attributes={
                'user_id': 'test_user',
                'project_id': 'test_project',
            },
            patchable_attributes={
                'host': 'foo',
                'image_ref': 'imageref!',
                'flavor_id': 1234,
                'display_name': 'myinstance',
            },
            resource_type='instance')),
        ('hardware.ipmi.node.power', dict(
            sample=sample.Sample(
                resource_id=str(uuid.uuid4()) + "_foobar",
                name='hardware.ipmi.node.power',
                unit='W',
                type=sample.TYPE_GAUGE,
                volume=2,
                user_id='test_user',
                project_id='test_project',
                source='openstack',
                timestamp='2012-05-08 20:23:48.028195',
                resource_metadata={
                    'useless': 'not_used',
                },
            ),
            metric_attributes={
                "archive_policy_name": "ceilometer-low",
                "unit": "W",
                "measures": [{
                    'timestamp': '2012-05-08 20:23:48.028195',
                    'value': 2
                }]
            },
            postable_attributes={
                'user_id': 'test_user',
                'project_id': 'test_project',
            },
            patchable_attributes={
            },
            resource_type='ipmi')),
    ]

    default_workflow = dict(resource_exists=True,
                            post_measure_fail=False,
                            create_resource_fail=False,
                            create_resource_race=False,
                            update_resource_fail=False,
                            retry_post_measures_fail=False)
    workflow_scenarios = [
        ('normal_workflow', {}),
        ('new_resource', dict(resource_exists=False)),
        ('new_resource_compat', dict(resource_exists=False)),
        ('new_resource_fail', dict(resource_exists=False,
                                   create_resource_fail=True)),
        ('new_resource_race', dict(resource_exists=False,
                                   create_resource_race=True)),
        ('resource_update_fail', dict(update_resource_fail=True)),
        ('retry_fail', dict(resource_exists=False,
                            retry_post_measures_fail=True)),
        ('measure_fail', dict(post_measure_fail=True)),
    ]

    @classmethod
    def generate_scenarios(cls):
        workflow_scenarios = []
        for name, wf_change in cls.workflow_scenarios:
            wf = cls.default_workflow.copy()
            wf.update(wf_change)
            workflow_scenarios.append((name, wf))
        cls.scenarios = testscenarios.multiply_scenarios(cls.sample_scenarios,
                                                         workflow_scenarios)

    def setUp(self):
        super().setUp()
        conf = ceilometer_service.prepare_service(argv=[], config_files=[])
        self.conf = self.useFixture(config_fixture.Config(conf))
        ks_client = mock.Mock()
        ks_client.projects.find.return_value = mock.Mock(
            name='gnocchi', id='a2d42c23-d518-46b6-96ab-3fba2e146859')
        self.useFixture(fixtures.MockPatch(
            'ceilometer.keystone_client.get_client',
            return_value=ks_client))
        self.useFixture(fixtures.MockPatch(
            'ceilometer.keystone_client.get_session',
            return_value=ks_client))
        self.ks_client = ks_client

    @mock.patch('gnocchiclient.v1.client.Client')
    def test_delete_event_workflow(self, fakeclient_cls):
        url = netutils.urlsplit("gnocchi://")
        self.publisher = gnocchi.GnocchiPublisher(self.conf.conf, url)

        fakeclient = fakeclient_cls.return_value

        fakeclient.resource.search.side_effect = [
            [{"id": "b26268d6-8bb5-11e6-baff-00224d8226cd",
              "type": "instance_disk",
              "instance_id": "9f9d01b9-4a58-4271-9e27-398b21ab20d1"}],
            [{"id": "b1c7544a-8bb5-11e6-850e-00224d8226cd",
              "type": "instance_network_interface",
              "instance_id": "9f9d01b9-4a58-4271-9e27-398b21ab20d1"}],
        ]

        search_params = {
            '=': {'instance_id': '9f9d01b9-4a58-4271-9e27-398b21ab20d1'}
        }

        now = timeutils.utcnow()
        self.useFixture(utils_fixture.TimeFixture(now))

        expected_calls = [
            mock.call.resource.search('instance_network_interface',
                                      search_params),
            mock.call.resource.search('instance_disk', search_params),
            mock.call.resource.update(
                'instance', '9f9d01b9-4a58-4271-9e27-398b21ab20d1',
                {'ended_at': now.isoformat()}),
            mock.call.resource.update(
                'instance_disk',
                'b26268d6-8bb5-11e6-baff-00224d8226cd',
                {'ended_at': now.isoformat()}),
            mock.call.resource.update(
                'instance_network_interface',
                'b1c7544a-8bb5-11e6-850e-00224d8226cd',
                {'ended_at': now.isoformat()}),
            mock.call.resource.update(
                'image', 'dc337359-de70-4044-8e2c-80573ba6e577',
                {'ended_at': now.isoformat()}),
            mock.call.resource.update(
                'volume', '6cc6e7dd-d17d-460f-ae79-7e08a216ce96',
                {'ended_at': now.isoformat()}),
            mock.call.resource.update(
                'network', '705e2c08-08e8-45cb-8673-5c5be955569b',
                {'ended_at': now.isoformat()})
        ]

        self.publisher.publish_events([INSTANCE_DELETE_START,
                                       IMAGE_DELETE_START,
                                       VOLUME_DELETE_END,
                                       FLOATINGIP_DELETE_END])
        self.assertEqual(8, len(fakeclient.mock_calls))
        for call in expected_calls:
            self.assertIn(call, fakeclient.mock_calls)

    @mock.patch('gnocchiclient.v1.client.Client')
    def test_create_event_workflow(self, fakeclient_cls):
        url = netutils.urlsplit("gnocchi://")
        self.publisher = gnocchi.GnocchiPublisher(self.conf.conf, url)

        fakeclient = fakeclient_cls.return_value

        now = timeutils.utcnow()
        self.useFixture(utils_fixture.TimeFixture(now))

        expected_calls = [
            mock.call.resource.create(
                'instance',
                {'id': '9f9d01b9-4a58-4271-9e27-398b21ab20d1',
                 'user_id': '1e3ce043029547f1a61c1996d1a531a2',
                 'project_id': '7c150a59fe714e6f9263774af9688f0e',
                 'availability_zone': 'zone1',
                 'flavor_name': 'm1.tiny',
                 'flavor_id': '2',
                 'host': 'vagrant-precise'}),
        ]

        self.publisher.publish_events([INSTANCE_CREATE_END])
        self.assertEqual(1, len(fakeclient.mock_calls))
        for call in expected_calls:
            self.assertIn(call, fakeclient.mock_calls)

    @mock.patch('gnocchiclient.v1.client.Client')
    def test_update_event_workflow(self, fakeclient_cls):
        url = netutils.urlsplit("gnocchi://")
        self.publisher = gnocchi.GnocchiPublisher(self.conf.conf, url)

        fakeclient = fakeclient_cls.return_value

        now = timeutils.utcnow()
        self.useFixture(utils_fixture.TimeFixture(now))

        expected_calls = [
            mock.call.resource.update(
                'volume',
                '156b8d3f-ad99-429b-b84c-3f263fb2a801',
                {'project_id': '85bc015f7a2342348593077a927c4aaa'}),
        ]

        self.publisher.publish_events([VOLUME_TRANSFER_ACCEPT_END])
        self.assertEqual(1, len(fakeclient.mock_calls))
        for call in expected_calls:
            self.assertIn(call, fakeclient.mock_calls)

    @mock.patch('gnocchiclient.v1.client.Client')
    def test_update_snapshot_event_workflow(self, fakeclient_cls):
        url = netutils.urlsplit("gnocchi://")
        self.publisher = gnocchi.GnocchiPublisher(self.conf.conf, url)

        fakeclient = fakeclient_cls.return_value

        now = timeutils.utcnow()
        self.useFixture(utils_fixture.TimeFixture(now))

        expected_calls = [
            mock.call.resource.update(
                'volume',
                '156b8d3f-ad99-429b-b84c-3f263fb2a801',
                {'project_id': '85bc015f7a2342348593077a927c4aaa'}),
        ]

        self.publisher.publish_events([SNAPSHOT_TRANSFER_ACCEPT_END])
        self.assertEqual(1, len(fakeclient.mock_calls))
        for call in expected_calls:
            self.assertIn(call, fakeclient.mock_calls)

    @mock.patch('ceilometer.cache_utils.get_client', mock.Mock())
    @mock.patch('ceilometer.publisher.gnocchi.LOG')
    @mock.patch('gnocchiclient.v1.client.Client')
    def test_workflow(self, fakeclient_cls, logger):
        url = netutils.urlsplit("gnocchi://")
        publisher = gnocchi.GnocchiPublisher(self.conf.conf, url)

        fakeclient = fakeclient_cls.return_value

        resource_id = self.sample.resource_id.replace("/", "_")
        metric_name = self.sample.name
        gnocchi_id = uuid.uuid4()

        expected_calls = [
            mock.call.archive_policy.create({"name": "ceilometer-low",
                                             "back_window": 0,
                                             "aggregation_methods": ["mean"],
                                             "definition": mock.ANY}),
            mock.call.archive_policy.create({"name": "ceilometer-low-rate",
                                             "back_window": 0,
                                             "aggregation_methods": [
                                                 "mean", "rate:mean"],
                                             "definition": mock.ANY}),
            mock.call.archive_policy.create({"name": "ceilometer-high",
                                             "back_window": 0,
                                             "aggregation_methods": ["mean"],
                                             "definition": mock.ANY}),
            mock.call.archive_policy.create({"name": "ceilometer-high-rate",
                                             "back_window": 0,
                                             "aggregation_methods": [
                                                 "mean", "rate:mean"],
                                             "definition": mock.ANY}),
            mock.call.metric.batch_resources_metrics_measures(
                {resource_id: {metric_name: self.metric_attributes}},
                create_metrics=True)
        ]

        resource_definition = publisher.metric_map.get(self.sample.name)

        expected_measures_in_log = {resource_id: {self.sample.name: {
            'measures': [{'timestamp': self.sample.timestamp,
                          'value': self.sample.volume}],
            'archive_policy_name': resource_definition.metrics[
                metric_name]["archive_policy_name"],
            'unit': self.sample.unit}}}

        resource_type = resource_definition.cfg['resource_type']

        expected_debug = [
            mock.call('Filtered project [%s] found with ID [%s].', 'service',
                      'a2d42c23-d518-46b6-96ab-3fba2e146859'),
            mock.call('Sample [%s] is not a Gnocchi activity; therefore, we '
                      'do not filter it out and push it to Gnocchi.',
                      self.sample),
            mock.call('Processing sample [%s] for resource ID [%s].',
                      self.sample, resource_id),
            mock.call('Executing batch resource metrics measures for resource '
                      '[%s] and measures [%s].',
                      mock.ANY,
                      expected_measures_in_log)]

        measures_posted = False
        batch_side_effect = []
        if self.post_measure_fail:
            batch_side_effect += [Exception('boom!')]
        elif not self.resource_exists:
            batch_side_effect += [
                gnocchi_exc.BadRequest(
                    400, {"cause": "Unknown resources",
                          'detail': [{
                              'resource_id': gnocchi_id,
                              'original_resource_id': resource_id}]})]

            attributes = self.postable_attributes.copy()
            attributes.update(self.patchable_attributes)
            attributes['id'] = self.sample.resource_id
            expected_calls.append(mock.call.resource.create(
                self.resource_type, attributes))

            if self.create_resource_fail:
                fakeclient.resource.create.side_effect = [Exception('boom!')]
            elif self.create_resource_race:
                fakeclient.resource.create.side_effect = [
                    gnocchi_exc.ResourceAlreadyExists(409)]
            else:  # not resource_exists
                expected_debug.append(mock.call(
                    'Resource %s created', self.sample.resource_id))

            if not self.create_resource_fail:
                expected_calls.append(
                    mock.call.metric.batch_resources_metrics_measures(
                        {resource_id: {metric_name: self.metric_attributes}},
                        create_metrics=True)
                )

                if self.retry_post_measures_fail:
                    batch_side_effect += [Exception('boom!')]
                else:
                    measures_posted = True

        else:
            measures_posted = True

        if measures_posted:
            batch_side_effect += [None]
            expected_debug.append(
                mock.call("%d measures posted against %d metrics through %d "
                          "resources", len(self.metric_attributes["measures"]),
                          1, 1)
            )

        if self.patchable_attributes:
            expected_calls.append(mock.call.resource.update(
                self.resource_type, resource_id,
                self.patchable_attributes))
            if self.update_resource_fail:
                fakeclient.resource.update.side_effect = [Exception('boom!')]
            else:
                expected_debug.append(mock.call(
                    'Resource %s updated', self.sample.resource_id))

        batch = fakeclient.metric.batch_resources_metrics_measures
        batch.side_effect = batch_side_effect

        publisher.publish_samples([self.sample])

        # Check that the last log message is the expected one
        if (self.post_measure_fail
                or self.create_resource_fail
                or self.retry_post_measures_fail
                or (self.update_resource_fail and self.patchable_attributes)):

            if self.update_resource_fail and self.patchable_attributes:
                logger.error.assert_called_with(
                    'Unexpected exception updating resource type [%s] with '
                    'ID [%s] for resource data [%s]: [%s].', resource_type,
                    resource_id, mock.ANY, 'boom!', exc_info=True)
            else:
                logger.error.assert_called_with(
                    'Unexpected exception while pushing measures [%s] for '
                    'gnocchi data [%s]: [%s].', expected_measures_in_log,
                    mock.ANY, 'boom!', exc_info=True)
        else:
            self.assertEqual(0, logger.error.call_count)
        self.assertEqual(expected_calls, fakeclient.mock_calls)
        self.assertEqual(expected_debug, logger.debug.mock_calls)


PublisherWorkflowTest.generate_scenarios()



#
# Copyright 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log
import tenacity

from ceilometer import service

LOG = log.getLogger(__name__)


def upgrade():
    conf = cfg.ConfigOpts()
    conf.register_cli_opts([
        cfg.BoolOpt('skip-gnocchi-resource-types',
                    help='Skip gnocchi resource-types upgrade.',
                    default=False),
        cfg.IntOpt('retry',
                   min=0,
                   help='Number of times to retry on failure. '
                   'Default is to retry forever.'),
    ])

    service.prepare_service(conf=conf)
    if conf.skip_gnocchi_resource_types:
        LOG.info("Skipping Gnocchi resource types upgrade")
    else:
        LOG.debug("Upgrading Gnocchi resource types")
        from ceilometer import gnocchi_client
        from gnocchiclient import exceptions
        if conf.retry is None:
            stop = tenacity.stop_never
        else:
            stop = tenacity.stop_after_attempt(conf.retry)
        tenacity.Retrying(
            stop=stop,
            retry=tenacity.retry_if_exception_type((
                exceptions.ConnectionFailure,
                exceptions.UnknownConnectionError,
                exceptions.ConnectionTimeout,
                exceptions.SSLError,
            ))
        )(gnocchi_client.upgrade_resource_types, conf)



# Copyright (C) 2026 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import firewall_group as sdk_firewall_group
from openstack.network.v2 import firewall_policy as sdk_firewall_policy
from openstack.network.v2 import floating_ip as sdk_floating_ip
from openstack.network.v2 import vpn_ipsec_site_connection as sdk_ipsec_conn
from openstack.network.v2 import vpn_service as sdk_vpn_service


class FakeSDKNetworkClient:

    def ips(self):
        FLOATING_IP_0 = sdk_floating_ip.FloatingIP(
            connection=None,
            id='fip-123',
            floating_ip_address='192.168.1.100',
            fixed_ip_address='10.0.0.5',
            status='ACTIVE',
            project_id='project-abc',
            router_id='router-456'
        )
        return iter([FLOATING_IP_0])

    def firewall_groups(self):
        FIREWALL_GROUP_0 = sdk_firewall_group.FirewallGroup(
            connection=None,
            id='fw-123',
            name='my-firewall',
            status='ACTIVE',
            project_id='project-abc',
            ingress_firewall_policy_id='policy-1',
            egress_firewall_policy_id='policy-2'
        )
        return iter([FIREWALL_GROUP_0])

    def firewall_policies(self):
        FIREWALL_POLICY_0 = sdk_firewall_policy.FirewallPolicy(
            connection=None,
            id='policy-123',
            name='my-policy',
            project_id='project-abc',
            firewall_rules=['rule-1', 'rule-2']
        )
        return iter([FIREWALL_POLICY_0])

    def vpn_ipsec_site_connections(self):
        VPN_IPSEC_CONN_0 = sdk_ipsec_conn.VpnIPSecSiteConnection(
            connection=None,
            id='ipsec-123',
            name='my-ipsec',
            status='ACTIVE',
            project_id='project-abc'
        )

        return iter([VPN_IPSEC_CONN_0])

    def vpn_services(self):

        VPN_SERVICE_0 = sdk_vpn_service.VpnService(
            connection=None,
            id='vpn-123',
            name='my-vpn',
            status='ACTIVE',
            project_id='project-abc'
        )

        return iter([VPN_SERVICE_0])


class FakeConnection:
    """Fake connection object for testing."""

    def __init__(self):
        """Initialize with a mock network attribute."""
        self.network = FakeSDKNetworkClient()



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Publish a sample using the preferred RPC mechanism.
"""

import abc
import itertools
import operator
import threading

from oslo_config import cfg
from oslo_log import log
import oslo_messaging
from oslo_utils import excutils
from urllib import parse as urlparse

from ceilometer import messaging
from ceilometer import publisher
from ceilometer.publisher import utils


LOG = log.getLogger(__name__)

NOTIFIER_OPTS = [
    cfg.StrOpt('metering_topic',
               default='metering',
               help='The topic that ceilometer uses for metering '
               'notifications.',
               deprecated_for_removal=True,
               ),
    cfg.StrOpt('event_topic',
               default='event',
               help='The topic that ceilometer uses for event '
               'notifications.',
               deprecated_for_removal=True,
               ),
    cfg.StrOpt('telemetry_driver',
               default='messagingv2',
               help='The driver that ceilometer uses for metering '
               'notifications.',
               deprecated_name='metering_driver',
               )
]


class DeliveryFailure(Exception):
    def __init__(self, message=None, cause=None):
        super().__init__(message)
        self.cause = cause


def raise_delivery_failure(exc):
    excutils.raise_with_cause(DeliveryFailure, str(exc), cause=exc)


class MessagingPublisher(publisher.ConfigPublisherBase, metaclass=abc.ABCMeta):

    def __init__(self, conf, parsed_url):
        super().__init__(conf, parsed_url)
        options = urlparse.parse_qs(parsed_url.query)
        # the value of options is a list of url param values
        # only take care of the latest one if the option
        # is provided more than once
        self.per_meter_topic = bool(int(
            options.get('per_meter_topic', [0])[-1]))

        self.policy = options.get('policy', ['default'])[-1]
        self.max_queue_length = int(options.get(
            'max_queue_length', [1024])[-1])
        self.max_retry = 0

        self.queue_lock = threading.Lock()
        self.local_queue = []

        if self.policy in ['default', 'queue', 'drop']:
            LOG.info('Publishing policy set to %s', self.policy)
        else:
            LOG.warning('Publishing policy is unknown (%s) force to default',
                        self.policy)
            self.policy = 'default'

        self.retry = 1 if self.policy in ['queue', 'drop'] else None

    def publish_samples(self, samples):
        """Publish samples on RPC.

        :param samples: Samples from pipeline.

        """

        meters = [
            utils.meter_message_from_counter(
                sample, self.conf.publisher.telemetry_secret)
            for sample in samples
        ]
        topic = self.conf.publisher_notifier.metering_topic
        with self.queue_lock:
            self.local_queue.append((topic, meters))

        if self.per_meter_topic:
            queue_per_meter_topic = []
            for meter_name, meter_list in itertools.groupby(
                    sorted(meters, key=operator.itemgetter('counter_name')),
                    operator.itemgetter('counter_name')):
                meter_list = list(meter_list)
                topic_name = topic + '.' + meter_name
                LOG.debug('Publishing %(m)d samples on %(n)s',
                          {'m': len(meter_list), 'n': topic_name})
                queue_per_meter_topic.append((topic_name, meter_list))
            with self.queue_lock:
                self.local_queue.extend(queue_per_meter_topic)

        self.flush()

    def flush(self):
        with self.queue_lock:
            queue = self.local_queue
            self.local_queue = []

        queue = self._process_queue(queue, self.policy)

        with self.queue_lock:
            self.local_queue = (queue + self.local_queue)
            if self.policy == 'queue':
                self._check_queue_length()

    def _check_queue_length(self):
        queue_length = len(self.local_queue)
        if queue_length > self.max_queue_length > 0:
            count = queue_length - self.max_queue_length
            self.local_queue = self.local_queue[count:]
            LOG.warning("Publisher max local_queue length is exceeded, "
                        "dropping %d oldest samples", count)

    def _process_queue(self, queue, policy):
        current_retry = 0
        while queue:
            topic, data = queue[0]
            try:
                self._send(topic, data)
            except DeliveryFailure:
                data = sum([len(m) for __, m in queue])
                if policy == 'queue':
                    LOG.warning("Failed to publish %d datapoints, queue them",
                                data)
                    return queue
                elif policy == 'drop':
                    LOG.warning("Failed to publish %d datapoints, "
                                "dropping them", data)
                    return []
                current_retry += 1
                if current_retry >= self.max_retry:
                    LOG.exception("Failed to retry to send sample data "
                                  "with max_retry times")
                    raise
            else:
                queue.pop(0)
        return []

    def publish_events(self, events):
        """Send an event message for publishing

        :param events: events from pipeline.
        """
        ev_list = [utils.message_from_event(
            event, self.conf.publisher.telemetry_secret) for event in events]

        topic = self.conf.publisher_notifier.event_topic
        with self.queue_lock:
            self.local_queue.append((topic, ev_list))
        self.flush()

    @abc.abstractmethod
    def _send(self, topic, meters):
        """Send the meters to the messaging topic."""


class NotifierPublisher(MessagingPublisher):
    """Publish metering data from notifier publisher.

    The ip address and port number of notifier can be configured in
    ceilometer pipeline configuration file.

    User can customize the transport driver such as rabbit, kafka and
    so on. The Notifier uses `sample` method as default method to send
    notifications.

    This publisher has transmit options such as queue, drop, and
    retry. These options are specified using policy field of URL parameter.
    When queue option could be selected, local queue length can be determined
    using max_queue_length field as well. When the transfer fails with retry
    option, try to resend the data as many times as specified in max_retry
    field. If max_retry is not specified, by default the number of retry
    is 100.

    To enable this publisher, add the following section to the
    /etc/ceilometer/pipeline.yaml file or simply add it to an existing
    pipeline::

        meter:
            - name: meter_notifier
              meters:
                - "*"
              sinks:
                - notifier_sink
        sinks:
            - name: notifier_sink
              publishers:
                - notifier://[notifier_ip]:[notifier_port]?topic=[topic]&
                  driver=driver&max_retry=100

    """

    def __init__(self, conf, parsed_url, default_topic):
        super().__init__(conf, parsed_url)
        options = urlparse.parse_qs(parsed_url.query)
        topics = options.pop('topic', [default_topic])
        driver = options.pop('driver', ['rabbit'])[0]
        self.max_retry = int(options.get('max_retry', [100])[-1])

        url = None
        if parsed_url.netloc != '':
            url = urlparse.urlunsplit([driver, parsed_url.netloc,
                                       parsed_url.path,
                                       urlparse.urlencode(options, True),
                                       parsed_url.fragment])
        self.notifier = oslo_messaging.Notifier(
            messaging.get_transport(self.conf, url),
            driver=self.conf.publisher_notifier.telemetry_driver,
            publisher_id='telemetry.publisher.%s' % self.conf.host,
            topics=topics,
            retry=self.retry
        )

    def _send(self, event_type, data):
        try:
            self.notifier.sample({}, event_type=event_type,
                                 payload=data)
        except oslo_messaging.MessageDeliveryFailure as e:
            raise_delivery_failure(e)


class SampleNotifierPublisher(NotifierPublisher):
    def __init__(self, conf, parsed_url):
        super().__init__(
            conf, parsed_url, conf.publisher_notifier.metering_topic)


class EventNotifierPublisher(NotifierPublisher):
    def __init__(self, conf, parsed_url):
        super().__init__(
            conf, parsed_url, conf.publisher_notifier.event_topic)



#
# Copyright 2024 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Tests for ceilometer polling heartbeat process"""

import multiprocessing
import shutil
import tempfile

from oslo_utils import timeutils
from unittest import mock

from ceilometer.polling import manager
from ceilometer import service
from ceilometer.tests import base


class TestHeartBeatManagert(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.conf = service.prepare_service([], [])
        self.tmpdir = tempfile.mkdtemp()

        self.queue = multiprocessing.Queue()
        self.mgr = manager.AgentManager(0, self.conf, namespaces='central',
                                        queue=self.queue)

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(self.tmpdir)

    def test_hb_not_configured(self):
        self.assertRaises(manager.HeartBeatException,
                          manager.AgentHeartBeatManager,
                          0, self.conf,
                          namespaces='ipmi',
                          queue=self.queue)

    @mock.patch('ceilometer.polling.manager.LOG')
    def test_hb_startup(self, LOG):
        # activate heartbeat agent
        self.conf.set_override('heartbeat_socket_dir', self.tmpdir,
                               group='polling')
        manager.AgentHeartBeatManager(0, self.conf, namespaces='compute',
                                      queue=self.queue)
        calls = [mock.call("Starting heartbeat child service. Listening"
                           f" on {self.tmpdir}/ceilometer-compute.socket")]
        LOG.info.assert_has_calls(calls)

    @mock.patch('ceilometer.polling.manager.LOG')
    def test_hb_update(self, LOG):
        self.conf.set_override('heartbeat_socket_dir', self.tmpdir,
                               group='polling')
        hb = manager.AgentHeartBeatManager(0, self.conf, namespaces='central',
                                           queue=self.queue)

        timestamp = timeutils.utcnow().isoformat()
        self.queue.put_nowait({'timestamp': timestamp, 'pollster': 'test'})

        hb._update_status()
        calls = [mock.call(f"Updated heartbeat for test ({timestamp})")]
        LOG.debug.assert_has_calls(calls)

    @mock.patch('ceilometer.polling.manager.LOG')
    def test_hb_send(self, LOG):
        with mock.patch('socket.socket') as FakeSocket:
            sub_skt = mock.Mock()
            sub_skt.sendall.return_value = None
            sub_skt.sendall.return_value = None

            skt = FakeSocket.return_value
            skt.bind.return_value = mock.Mock()
            skt.listen.return_value = mock.Mock()
            skt.accept.return_value = (sub_skt, "")

            self.conf.set_override('heartbeat_socket_dir', self.tmpdir,
                                   group='polling')
            hb = manager.AgentHeartBeatManager(0, self.conf,
                                               namespaces='central',
                                               queue=self.queue)
            timestamp = timeutils.utcnow().isoformat()
            self.queue.put_nowait({'timestamp': timestamp,
                                   'pollster': 'test1'})
            hb._update_status()
            self.queue.put_nowait({'timestamp': timestamp,
                                   'pollster': 'test2'})
            hb._update_status()

            # test status report
            hb._send_heartbeat()
            calls = [mock.call("Heartbeat status report requested "
                               f"at {self.tmpdir}/ceilometer-central.socket"),
                     mock.call("Reported heartbeat status:\n"
                               f"test1 {timestamp}\n"
                               f"test2 {timestamp}")]
            LOG.debug.assert_has_calls(calls)



#
# Copyright 2012 eNovance <licensing@enovance.com>
# Copyright 2012 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ceilometer.compute import pollsters
from ceilometer.compute.pollsters import util
from ceilometer import sample


class NetworkPollster(pollsters.GenericComputePollster):
    inspector_method = "inspect_vnics"

    @staticmethod
    def get_additional_metadata(instance, stats):
        additional_stats = {k: getattr(stats, k)
                            for k in ["name", "mac", "fref", "parameters"]}
        if stats.fref is not None:
            additional_stats['vnic_name'] = stats.fref
        else:
            additional_stats['vnic_name'] = stats.name
        return additional_stats

    @staticmethod
    def get_resource_id(instance, stats):
        if stats.fref is not None:
            return stats.fref
        else:
            instance_name = util.instance_name(instance)
            return f"{instance_name}-{instance.id}-{stats.name}"


class IncomingBytesPollster(NetworkPollster):
    sample_name = 'network.incoming.bytes'
    sample_type = sample.TYPE_CUMULATIVE
    sample_unit = 'B'
    sample_stats_key = 'rx_bytes'


class IncomingPacketsPollster(NetworkPollster):
    sample_name = 'network.incoming.packets'
    sample_type = sample.TYPE_CUMULATIVE
    sample_unit = 'packet'
    sample_stats_key = 'rx_packets'


class OutgoingBytesPollster(NetworkPollster):
    sample_name = 'network.outgoing.bytes'
    sample_type = sample.TYPE_CUMULATIVE
    sample_unit = 'B'
    sample_stats_key = 'tx_bytes'


class OutgoingPacketsPollster(NetworkPollster):
    sample_name = 'network.outgoing.packets'
    sample_type = sample.TYPE_CUMULATIVE
    sample_unit = 'packet'
    sample_stats_key = 'tx_packets'


class IncomingBytesRatePollster(NetworkPollster):
    inspector_method = "inspect_vnic_rates"
    sample_name = 'network.incoming.bytes.rate'
    sample_unit = 'B/s'
    sample_stats_key = 'rx_bytes_rate'


class OutgoingBytesRatePollster(NetworkPollster):
    inspector_method = "inspect_vnic_rates"
    sample_name = 'network.outgoing.bytes.rate'
    sample_unit = 'B/s'
    sample_stats_key = 'tx_bytes_rate'


class IncomingDropPollster(NetworkPollster):
    sample_name = 'network.incoming.packets.drop'
    sample_type = sample.TYPE_CUMULATIVE
    sample_unit = 'packet'
    sample_stats_key = 'rx_drop'


class OutgoingDropPollster(NetworkPollster):
    sample_name = 'network.outgoing.packets.drop'
    sample_type = sample.TYPE_CUMULATIVE
    sample_unit = 'packet'
    sample_stats_key = 'tx_drop'


class IncomingErrorsPollster(NetworkPollster):
    sample_name = 'network.incoming.packets.error'
    sample_type = sample.TYPE_CUMULATIVE
    sample_unit = 'packet'
    sample_stats_key = 'rx_errors'


class OutgoingErrorsPollster(NetworkPollster):
    sample_name = 'network.outgoing.packets.error'
    sample_type = sample.TYPE_CUMULATIVE
    sample_unit = 'packet'
    sample_stats_key = 'tx_errors'


class IncomingBytesDeltaPollster(NetworkPollster):
    sample_name = 'network.incoming.bytes.delta'
    sample_type = sample.TYPE_DELTA
    sample_unit = 'B'
    sample_stats_key = 'rx_bytes_delta'


class OutgoingBytesDeltaPollster(NetworkPollster):
    sample_name = 'network.outgoing.bytes.delta'
    sample_type = sample.TYPE_DELTA
    sample_unit = 'B'
    sample_stats_key = 'tx_bytes_delta'



# Copyright (C) 2026 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack import exceptions as os_exc
from oslotest import base

from ceilometer import neutron_client
from ceilometer import service
from ceilometer.tests.unit import fakes


class TestNeutronClient(base.BaseTestCase):
    """Tests for the Neutron client using openstacksdk."""

    def setUp(self):
        super().setUp()
        self.CONF = service.prepare_service([], [])

        self.mock_connection = mock.patch(
            'openstack.connection.Connection',
            autospec=True, return_value=fakes.FakeConnection())
        self.mock_conn_class = self.mock_connection.start()
        self.addCleanup(self.mock_connection.stop)

        # Create the client (uses mocked Connection)
        self.nc = neutron_client.Client(self.CONF)

    def test_fip_get_all(self):
        """Test getting all floating IPs returns list of dicts."""
        result = self.nc.fip_get_all()

        self.assertEqual(1, len(result))
        self.assertIsInstance(result[0], dict)
        self.assertEqual('fip-123', result[0]['id'])
        self.assertEqual('192.168.1.100', result[0]['floating_ip_address'])

    def test_fip_get_all_empty(self):
        """Test getting floating IPs when none exist."""
        self.nc.conn.network.ips = mock.Mock(return_value=iter([]))

        result = self.nc.fip_get_all()

        self.assertEqual([], result)

    def test_vpn_get_all(self):
        """Test getting all VPN services returns list of dicts."""

        result = self.nc.vpn_get_all()

        self.assertEqual(1, len(result))
        self.assertIsInstance(result[0], dict)
        self.assertEqual('vpn-123', result[0]['id'])
        self.assertEqual('my-vpn', result[0]['name'])

    def test_vpn_get_all_empty(self):
        """Test getting VPN services when none exist."""
        self.nc.conn.network.vpn_services = mock.Mock(
            return_value=iter([]))

        result = self.nc.vpn_get_all()

        self.assertEqual([], result)

    def test_ipsec_site_connections_get_all(self):
        """Test getting all IPSec site connections returns list of dicts."""

        result = self.nc.ipsec_site_connections_get_all()

        self.assertEqual(1, len(result))
        self.assertIsInstance(result[0], dict)
        self.assertEqual('ipsec-123', result[0]['id'])

    def test_ipsec_site_connections_get_all_empty(self):
        """Test getting IPSec connections when none exist."""
        self.nc.conn.network.vpn_ipsec_site_connections =\
            mock.Mock(return_value=iter([]))

        result = self.nc.ipsec_site_connections_get_all()

        self.assertEqual([], result)

    def test_firewall_get_all(self):
        """Test getting all firewall groups returns list of dicts."""

        result = self.nc.firewall_get_all()

        self.assertEqual(1, len(result))
        self.assertIsInstance(result[0], dict)
        self.assertEqual('fw-123', result[0]['id'])
        self.assertEqual('my-firewall', result[0]['name'])

    def test_firewall_get_all_empty(self):
        """Test getting firewall groups when none exist."""
        self.nc.conn.network.firewall_groups = mock.Mock(
            return_value=iter([]))

        result = self.nc.firewall_get_all()

        self.assertEqual([], result)

    def test_fw_policy_get_all(self):
        """Test getting all firewall policies returns list of dicts."""

        result = self.nc.fw_policy_get_all()

        self.assertEqual(1, len(result))
        self.assertIsInstance(result[0], dict)
        self.assertEqual('policy-123', result[0]['id'])
        self.assertEqual('my-policy', result[0]['name'])

    def test_fw_policy_get_all_empty(self):
        """Test getting firewall policies when none exist."""
        self.nc.conn.network.firewall_policies = mock.Mock(
            return_value=iter([]))

        result = self.nc.fw_policy_get_all()

        self.assertEqual([], result)

    def test_fip_get_all_404_returns_empty(self):
        """Test that 404 errors return empty list."""
        self.nc.conn.network.ips = mock.Mock(
            side_effect=os_exc.HttpException(http_status=404))

        result = self.nc.fip_get_all()

        self.assertEqual([], result)

    def test_fip_get_all_other_http_error_returns_empty(self):
        """Test that other HTTP errors return empty list."""
        self.nc.conn.network.ips = mock.Mock(side_effect=os_exc.HttpException(
            http_status=500))

        result = self.nc.fip_get_all()

        self.assertEqual([], result)

    def test_fip_get_all_non_http_exception_raises(self):
        """Test that non-HTTP exceptions are re-raised."""
        self.nc.conn.network.ips = mock.Mock(side_effect=RuntimeError(
            "Connection failed"))

        self.assertRaises(RuntimeError, self.nc.fip_get_all)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Tests for ceilometer.meter.notifications"""
import copy
from unittest import mock

import fixtures
from oslo_cache import core as cache
from oslo_config import fixture as config_fixture
from oslo_utils import fileutils
import yaml

from ceilometer import declarative
from ceilometer.meter import notifications
from ceilometer import service as ceilometer_service
from ceilometer.tests import base as test

NOTIFICATION = {
    'event_type': 'test.create',
    'metadata': {'timestamp': '2015-06-19T09:19:35.786893',
                 'message_id': '939823de-c242-45a2-a399-083f4d6a8c3e'},
    'payload': {'user_id': 'e1d870e51c7340cb9d555b15cbfcaec2',
                'resource_id': 'bea70e51c7340cb9d555b15cbfcaec23',
                'timestamp': '2015-06-19T09:19:35.785330',
                'created_at': '2015-06-19T09:25:35.785330',
                'launched_at': '2015-06-19T09:25:40.785330',
                'message_signature': 'fake_signature1',
                'resource_metadata': {'foo': 'bar'},
                'source': '30be1fc9a03c4e94ab05c403a8a377f2: openstack',
                'volume': 1.0,
                'project_id': '30be1fc9a03c4e94ab05c403a8a377f2',
                },
    'ctxt': {'tenant': '30be1fc9a03c4e94ab05c403a8a377f2',
             'request_id': 'req-da91b4bf-d2b5-43ae-8b66-c7752e72726d',
             'user': 'e1d870e51c7340cb9d555b15cbfcaec2'},
    'publisher_id': "foo123"
}

USER_META = {
    'event_type': 'test.create',
    'metadata': {'timestamp': '2015-06-19T09:19:35.786893',
                 'message_id': '939823de-c242-45a2-a399-083f4d6a8c3e'},
    'payload': {'user_id': 'e1d870e51c7340cb9d555b15cbfcaec2',
                'resource_id': 'bea70e51c7340cb9d555b15cbfcaec23',
                'timestamp': '2015-06-19T09:19:35.785330',
                'created_at': '2015-06-19T09:25:35.785330',
                'launched_at': '2015-06-19T09:25:40.785330',
                'message_signature': 'fake_signature1',
                'resource_metadata': {'foo': 'bar'},
                'source': '30be1fc9a03c4e94ab05c403a8a377f2: openstack',
                'volume': 1.0,
                'project_id': '30be1fc9a03c4e94ab05c403a8a377f2',
                'metadata': {'metering.xyz': 'abc', 'ignore': 'this'},
                },
    'ctxt': {'tenant': '30be1fc9a03c4e94ab05c403a8a377f2',
             'request_id': 'req-da91b4bf-d2b5-43ae-8b66-c7752e72726d',
             'user': 'e1d870e51c7340cb9d555b15cbfcaec2'},
    'publisher_id': "foo123"
}

MIDDLEWARE_EVENT = {
    'ctxt': {'request_id': 'req-a8bfa89b-d28b-4b95-9e4b-7d7875275650',
             'quota_class': None,
             'service_catalog': [],
             'auth_token': None,
             'user_id': None,
             'is_admin': True,
             'user': None,
             'remote_address': None,
             'roles': [],
             'timestamp': '2013-07-29T06:51:34.348091',
             'project_name': None,
             'read_deleted': 'no',
             'tenant': None,
             'instance_lock_checked': False,
             'project_id': None,
             'user_name': None},
    'event_type': 'objectstore.http.request',
    'publisher_id': 'ceilometermiddleware',
    'metadata': {'message_id': '6eccedba-120e-4db8-9735-2ad5f061e5ee',
                  'timestamp': '2013-07-29T06:51:34.474815+00:00',
                  '_unique_id': '0ee26117077648e18d88ac76e28a72e2'},
    'payload': {
        'typeURI': 'http: //schemas.dmtf.org/cloud/audit/1.0/event',
        'eventTime': '2013-07-29T06:51:34.474815+00:00',
        'target': {
            'action': 'get',
            'typeURI': 'service/storage/object',
            'id': 'account',
            'metadata': {
                'path': '/1.0/CUSTOM_account/container/obj',
                'version': '1.0',
                'container': 'container',
                'object': 'obj'
            }
        },
        'observer': {
            'id': 'target'
        },
        'eventType': 'activity',
        'measurements': [
            {
                'metric': {
                    'metricId': 'openstack: uuid',
                    'name': 'storage.objects.outgoing.bytes',
                    'unit': 'B'
                },
                'result': 28
            },
            {
                'metric': {
                    'metricId': 'openstack: uuid2',
                    'name': 'storage.objects.incoming.bytes',
                    'unit': 'B'
                },
                'result': 1
            }
        ],
        'initiator': {
            'typeURI': 'service/security/account/user',
            'project_id': None,
            'id': 'openstack: 288f6260-bf37-4737-a178-5038c84ba244'
        },
        'action': 'read',
        'outcome': 'success',
        'id': 'openstack: 69972bb6-14dd-46e4-bdaf-3148014363dc'
    }
}

FULL_MULTI_MSG = {
    'event_type': 'full.sample',
    'payload': [{
                'counter_name': 'instance1',
                'user_id': 'user1',
                'user_name': 'fake-name',
                'resource_id': 'res1',
                'counter_unit': 'ns',
                'counter_volume': 28.0,
                'project_id': 'proj1',
                'project_name': 'fake-name',
                'counter_type': 'gauge'
                },
                {
                'counter_name': 'instance2',
                'user_id': 'user2',
                'user_name': 'fake-name',
                'resource_id': 'res2',
                'counter_unit': '%',
                'counter_volume': 1.0,
                'project_id': 'proj2',
                'project_name': 'fake-name',
                'counter_type': 'delta'
                }],
    'ctxt': {'domain': None,
             'request_id': 'req-da91b4bf-d2b5-43ae-8b66-c7752e72726d',
             'auth_token': None,
             'read_only': False,
             'resource_uuid': None,
             'user_identity': 'fake_user_identity---',
             'show_deleted': False,
             'tenant': '30be1fc9a03c4e94ab05c403a8a377f2',
             'is_admin': True,
             'project_domain': None,
             'user': 'e1d870e51c7340cb9d555b15cbfcaec2',
             'user_domain': None},
    'publisher_id': 'ceilometer.api',
    'metadata': {'message_id': '939823de-c242-45a2-a399-083f4d6a8c3e',
                 'timestamp': '2015-06-19T09:19:35.786893'},
}

METRICS_UPDATE = {
    'event_type': 'compute.metrics.update',
    'payload': {
        'metrics': [
            {'timestamp': '2013-07-29T06:51:34.472416',
             'name': 'cpu.frequency', 'value': 1600,
             'source': 'libvirt.LibvirtDriver'},
            {'timestamp': '2013-07-29T06:51:34.472416',
             'name': 'cpu.user.time', 'value': 17421440000000,
             'source': 'libvirt.LibvirtDriver'},
            {'timestamp': '2013-07-29T06:51:34.472416',
             'name': 'cpu.kernel.time', 'value': 7852600000000,
             'source': 'libvirt.LibvirtDriver'},
            {'timestamp': '2013-07-29T06:51:34.472416',
             'name': 'cpu.idle.time', 'value': 1307374400000000,
             'source': 'libvirt.LibvirtDriver'},
            {'timestamp': '2013-07-29T06:51:34.472416',
             'name': 'cpu.iowait.time', 'value': 11697470000000,
             'source': 'libvirt.LibvirtDriver'},
            {'timestamp': '2013-07-29T06:51:34.472416',
             'name': 'cpu.user.percent', 'value': 0.012959045637294348,
             'source': 'libvirt.LibvirtDriver'},
            {'timestamp': '2013-07-29T06:51:34.472416',
             'name': 'cpu.kernel.percent', 'value': 0.005841204961898534,
             'source': 'libvirt.LibvirtDriver'},
            {'timestamp': '2013-07-29T06:51:34.472416',
             'name': 'cpu.idle.percent', 'value': 0.9724985141658965,
             'source': 'libvirt.LibvirtDriver'},
            {'timestamp': '2013-07-29T06:51:34.472416',
             'name': 'cpu.iowait.percent', 'value': 0.008701235234910634,
             'source': 'libvirt.LibvirtDriver'},
            {'timestamp': '2013-07-29T06:51:34.472416',
             'name': 'cpu.percent', 'value': 0.027501485834103515,
             'source': 'libvirt.LibvirtDriver'}],
        'nodename': 'tianst.sh.intel.com',
        'host': 'tianst',
        'host_id': '10.0.1.1'},
    'publisher_id': 'compute.tianst.sh.intel.com',
    'metadata': {'message_id': '6eccedba-120e-4db8-9735-2ad5f061e5ee',
                 'timestamp': '2013-07-29 06:51:34.474815',
                 '_unique_id': '0ee26117077648e18d88ac76e28a72e2'},
    'ctxt': {'request_id': 'req-a8bfa89b-d28b-4b95-9e4b-7d7875275650',
             'quota_class': None,
             'service_catalog': [],
             'auth_token': None,
             'user_id': None,
             'is_admin': True,
             'user': None,
             'remote_address': None,
             'roles': [],
             'timestamp': '2013-07-29T06:51:34.348091',
             'project_name': None,
             'read_deleted': 'no',
             'tenant': None,
             'instance_lock_checked': False,
             'project_id': None,
             'user_name': None}
}


class TestMeterDefinition(test.BaseTestCase):
    def test_config_definition(self):
        cfg = dict(name="test",
                   event_type="test.create",
                   type="delta",
                   unit="B",
                   volume="$.payload.volume",
                   resource_id="$.payload.resource_id",
                   project_id="$.payload.project_id")
        conf = ceilometer_service.prepare_service([], [])
        handler = notifications.MeterDefinition(cfg, conf, mock.Mock())
        self.assertTrue(handler.match_type("test.create"))
        sample = list(handler.to_samples(NOTIFICATION))[0]
        self.assertEqual(1.0, sample["volume"])
        self.assertEqual("bea70e51c7340cb9d555b15cbfcaec23",
                         sample["resource_id"])
        self.assertEqual("30be1fc9a03c4e94ab05c403a8a377f2",
                         sample["project_id"])

    def test_config_required_missing_fields(self):
        cfg = dict()
        conf = ceilometer_service.prepare_service([], [])
        try:
            notifications.MeterDefinition(cfg, conf, mock.Mock())
        except declarative.DefinitionException as e:
            self.assertIn("Required fields ['name', 'type', 'event_type',"
                          " 'unit', 'volume', 'resource_id']"
                          " not specified",
                          str(e))

    def test_bad_type_cfg_definition(self):
        cfg = dict(name="test", type="foo", event_type="bar.create",
                   unit="foo", volume="bar",
                   resource_id="bea70e51c7340cb9d555b15cbfcaec23")
        conf = ceilometer_service.prepare_service([], [])
        try:
            notifications.MeterDefinition(cfg, conf, mock.Mock())
        except declarative.DefinitionException as e:
            self.assertIn("Invalid type foo specified", str(e))


class CacheConfFixture(config_fixture.Config):
    def setUp(self):
        super().setUp()
        self.conf = ceilometer_service.\
            prepare_service(argv=[], config_files=[])
        cache.configure(self.conf)


class TestMeterProcessing(test.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.CONF = ceilometer_service.prepare_service([], [])
        dict_conf_fixture = CacheConfFixture(self.CONF)
        self.useFixture(dict_conf_fixture)
        dict_conf_fixture.config(enabled=True, group='cache')
        dict_conf_fixture.config(expiration_time=600,
                                 backend='oslo_cache.dict',
                                 group='cache')
        dict_conf_fixture.config(identity_name_discovery=True, group='polling')
        self.CONF = dict_conf_fixture.conf

        self.path = self.useFixture(fixtures.TempDir()).path
        self.handler = notifications.ProcessMeterNotifications(
            self.CONF, mock.Mock())

    def _load_meter_def_file(self, cfgs=None):
        self.CONF.set_override('meter_definitions_dirs',
                               [self.path], group='meter')
        cfgs = cfgs or []
        if not isinstance(cfgs, list):
            cfgs = [cfgs]
        meter_cfg_files = list()
        for cfg in cfgs:
            cfg = cfg.encode('utf-8')
            meter_cfg_files.append(fileutils.write_to_tempfile(content=cfg,
                                                               path=self.path,
                                                               prefix="meters",
                                                               suffix=".yaml"))
        self.handler.definitions = self.handler._load_definitions()

    @mock.patch('ceilometer.meter.notifications.LOG')
    def test_bad_meter_definition_skip(self, LOG):
        cfg = yaml.dump(
            {'metric': [dict(name="good_test_1",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id"),
                        dict(name="bad_test_2", type="bad_type",
                             event_type="bar.create",
                             unit="foo", volume="bar",
                             resource_id="bea70e51c7340cb9d555b15cbfcaec23"),
                        dict(name="good_test_3",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        self._load_meter_def_file(cfg)
        self.assertEqual(2, len(self.handler.definitions))
        args, kwargs = LOG.error.call_args_list[0]
        self.assertEqual("Error loading meter definition: %s", args[0])
        self.assertTrue(
            str(args[1]).endswith("Invalid type bad_type specified"))

    def test_jsonpath_values_parsed(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                        event_type="test.create",
                        type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        self.assertEqual('test1', s1['name'])
        self.assertEqual(1.0, s1['volume'])
        self.assertEqual('bea70e51c7340cb9d555b15cbfcaec23', s1['resource_id'])
        self.assertEqual('30be1fc9a03c4e94ab05c403a8a377f2', s1['project_id'])

    def test_multiple_meter(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                        event_type="test.create",
                        type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id"),
                        dict(name="test2",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        self._load_meter_def_file(cfg)
        data = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(2, len(data))
        expected_names = ['test1', 'test2']
        for s in data:
            self.assertIn(s.as_dict()['name'], expected_names)

    def test_unmatched_meter(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                        event_type="test.update",
                        type="delta",
                        unit="B",
                        volume="$.payload.volume",
                        resource_id="$.payload.resource_id",
                        project_id="$.payload.project_id")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(0, len(c))

    def test_regex_match_meter(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                        event_type="test.*",
                        type="delta",
                        unit="B",
                        volume="$.payload.volume",
                        resource_id="$.payload.resource_id",
                        project_id="$.payload.project_id")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(1, len(c))

    def test_default_timestamp(self):
        event = copy.deepcopy(MIDDLEWARE_EVENT)
        del event['payload']['measurements'][1]
        cfg = yaml.dump(
            {'metric': [dict(name="$.payload.measurements.[*].metric.[*].name",
                        event_type="objectstore.http.request",
                        type="delta",
                        unit="$.payload.measurements.[*].metric.[*].unit",
                        volume="$.payload.measurements.[*].result",
                        resource_id="$.payload.target_id",
                        project_id="$.payload.initiator.project_id",
                        multi="name")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(event))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        self.assertEqual(MIDDLEWARE_EVENT['metadata']['timestamp'],
                         s1['timestamp'])

    def test_custom_timestamp(self):
        event = copy.deepcopy(MIDDLEWARE_EVENT)
        del event['payload']['measurements'][1]
        cfg = yaml.dump(
            {'metric': [dict(name="$.payload.measurements.[*].metric.[*].name",
                        event_type="objectstore.http.request",
                        type="delta",
                        unit="$.payload.measurements.[*].metric.[*].unit",
                        volume="$.payload.measurements.[*].result",
                        resource_id="$.payload.target_id",
                        project_id="$.payload.initiator.project_id",
                        multi="name",
                        timestamp='$.payload.eventTime')]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(event))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        self.assertEqual(MIDDLEWARE_EVENT['payload']['eventTime'],
                         s1['timestamp'])

    def test_custom_timestamp_expr_meter(self):
        cfg = yaml.dump(
            {'metric': [dict(name='compute.node.cpu.frequency',
                        event_type="compute.metrics.update",
                        type='gauge',
                        unit="ns",
                        volume="$.payload.metrics[?(@.name='cpu.frequency')]"
                               ".value",
                        resource_id="'prefix-' + $.payload.nodename",
                        timestamp="$.payload.metrics"
                                  "[?(@.name='cpu.frequency')].timestamp")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(METRICS_UPDATE))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        self.assertEqual('compute.node.cpu.frequency', s1['name'])
        self.assertEqual("2013-07-29T06:51:34.472416+00:00", s1['timestamp'])

    def test_default_metadata(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                        event_type="test.*",
                        type="delta",
                        unit="B",
                        volume="$.payload.volume",
                        resource_id="$.payload.resource_id",
                        project_id="$.payload.project_id")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        meta = NOTIFICATION['payload'].copy()
        meta['host'] = NOTIFICATION['publisher_id']
        meta['event_type'] = NOTIFICATION['event_type']
        self.assertEqual(meta, s1['resource_metadata'])

    def test_datetime_plugin(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                        event_type="test.*",
                        type="gauge",
                        unit="sec",
                        volume={"fields": ["$.payload.created_at",
                                           "$.payload.launched_at"],
                                "plugin": "timedelta"},
                        resource_id="$.payload.resource_id",
                        project_id="$.payload.project_id")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        self.assertEqual(5.0, s1['volume'])

    def test_custom_metadata(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                        event_type="test.*",
                        type="delta",
                        unit="B",
                        volume="$.payload.volume",
                        resource_id="$.payload.resource_id",
                        project_id="$.payload.project_id",
                        metadata={'proj': '$.payload.project_id',
                                  'dict': '$.payload.resource_metadata'})]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        meta = {'proj': s1['project_id'],
                'dict': NOTIFICATION['payload']['resource_metadata']}
        self.assertEqual(meta, s1['resource_metadata'])

    def test_user_meta(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                        event_type="test.*",
                        type="delta",
                        unit="B",
                        volume="$.payload.volume",
                        resource_id="$.payload.resource_id",
                        project_id="$.payload.project_id",
                        user_metadata="$.payload.metadata",)]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(USER_META))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        meta = {'user_metadata': {'xyz': 'abc'}}
        self.assertEqual(meta, s1['resource_metadata'])

    def test_user_meta_and_custom(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                        event_type="test.*",
                        type="delta",
                        unit="B",
                        volume="$.payload.volume",
                        resource_id="$.payload.resource_id",
                        project_id="$.payload.project_id",
                        user_metadata="$.payload.metadata",
                        metadata={'proj': '$.payload.project_id'})]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(USER_META))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        meta = {'user_metadata': {'xyz': 'abc'}, 'proj': s1['project_id']}
        self.assertEqual(meta, s1['resource_metadata'])

    def test_multi_match_event_meter(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                        event_type="test.create",
                        type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id"),
                        dict(name="test2",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(2, len(c))

    def test_multi_meter_payload(self):
        cfg = yaml.dump(
            {'metric': [dict(name="$.payload.measurements.[*].metric.[*].name",
                        event_type="objectstore.http.request",
                        type="delta",
                        unit="$.payload.measurements.[*].metric.[*].unit",
                        volume="$.payload.measurements.[*].result",
                        resource_id="$.payload.target_id",
                        project_id="$.payload.initiator.project_id",
                        lookup=["name", "volume", "unit"])]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(MIDDLEWARE_EVENT))
        self.assertEqual(2, len(c))
        s1 = c[0].as_dict()
        self.assertEqual('storage.objects.outgoing.bytes', s1['name'])
        self.assertEqual(28, s1['volume'])
        self.assertEqual('B', s1['unit'])
        s2 = c[1].as_dict()
        self.assertEqual('storage.objects.incoming.bytes', s2['name'])
        self.assertEqual(1, s2['volume'])
        self.assertEqual('B', s2['unit'])

    def test_multi_meter_payload_single(self):
        event = copy.deepcopy(MIDDLEWARE_EVENT)
        del event['payload']['measurements'][1]
        cfg = yaml.dump(
            {'metric': [dict(name="$.payload.measurements.[*].metric.[*].name",
                        event_type="objectstore.http.request",
                        type="delta",
                        unit="$.payload.measurements.[*].metric.[*].unit",
                        volume="$.payload.measurements.[*].result",
                        resource_id="$.payload.target_id",
                        project_id="$.payload.initiator.project_id",
                        lookup=["name", "unit"])]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(event))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        self.assertEqual('storage.objects.outgoing.bytes', s1['name'])
        self.assertEqual(28, s1['volume'])
        self.assertEqual('B', s1['unit'])

    def test_multi_meter_payload_none(self):
        event = copy.deepcopy(MIDDLEWARE_EVENT)
        del event['payload']['measurements']
        cfg = yaml.dump(
            {'metric': [dict(name="$.payload.measurements.[*].metric.[*].name",
                        event_type="objectstore.http.request",
                        type="delta",
                        unit="$.payload.measurements.[*].metric.[*].unit",
                        volume="$.payload.measurements.[*].result",
                        resource_id="$.payload.target_id",
                        project_id="$.payload.initiator.project_id",
                        lookup="name")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(event))
        self.assertEqual(0, len(c))

    @mock.patch(
        'ceilometer.cache_utils.CacheClient._resolve_uuid_from_keystone'
    )
    def test_multi_meter_payload_all_multi(self, resolved_uuid):
        resolved_uuid.return_value = "fake-name"
        cfg = yaml.dump(
            {'metric': [dict(name="$.payload.[*].counter_name",
                        event_type="full.sample",
                        type="$.payload.[*].counter_type",
                        unit="$.payload.[*].counter_unit",
                        volume="$.payload.[*].counter_volume",
                        resource_id="$.payload.[*].resource_id",
                        project_id="$.payload.[*].project_id",
                        user_id="$.payload.[*].user_id",
                        lookup=['name', 'type', 'unit', 'volume',
                                'resource_id', 'project_id', 'user_id'])]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(FULL_MULTI_MSG))
        self.assertEqual(2, len(c))
        msg = FULL_MULTI_MSG['payload']
        for idx, val in enumerate(c):
            s1 = val.as_dict()
            self.assertEqual(msg[idx]['counter_name'], s1['name'])
            self.assertEqual(msg[idx]['counter_volume'], s1['volume'])
            self.assertEqual(msg[idx]['counter_unit'], s1['unit'])
            self.assertEqual(msg[idx]['counter_type'], s1['type'])
            self.assertEqual(msg[idx]['resource_id'], s1['resource_id'])
            self.assertEqual(msg[idx]['project_id'], s1['project_id'])
            self.assertEqual(msg[idx]['user_id'], s1['user_id'])
            self.assertEqual(msg[idx]['project_name'], s1['project_name'])
            self.assertEqual(msg[idx]['user_name'], s1['user_name'])

    @mock.patch('ceilometer.meter.notifications.LOG')
    def test_multi_meter_payload_invalid_missing(self, LOG):
        event = copy.deepcopy(MIDDLEWARE_EVENT)
        del event['payload']['measurements'][0]['result']
        del event['payload']['measurements'][1]['result']
        cfg = yaml.dump(
            {'metric': [dict(name="$.payload.measurements.[*].metric.[*].name",
                        event_type="objectstore.http.request",
                        type="delta",
                        unit="$.payload.measurements.[*].metric.[*].unit",
                        volume="$.payload.measurements.[*].result",
                        resource_id="$.payload.target_id",
                        project_id="$.payload.initiator.project_id",
                        lookup=["name", "unit", "volume"])]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(event))
        self.assertEqual(0, len(c))
        log_called_args = LOG.warning.call_args_list
        self.assertEqual(
            'Only 0 fetched meters contain "volume" field instead of 2.',
            log_called_args[0][0][0] % log_called_args[0][0][1])

    @mock.patch('ceilometer.meter.notifications.LOG')
    def test_multi_meter_payload_invalid_short(self, LOG):
        event = copy.deepcopy(MIDDLEWARE_EVENT)
        del event['payload']['measurements'][0]['result']
        cfg = yaml.dump(
            {'metric': [dict(name="$.payload.measurements.[*].metric.[*].name",
                        event_type="objectstore.http.request",
                        type="delta",
                        unit="$.payload.measurements.[*].metric.[*].unit",
                        volume="$.payload.measurements.[*].result",
                        resource_id="$.payload.target_id",
                        project_id="$.payload.initiator.project_id",
                        lookup=["name", "unit", "volume"])]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(event))
        self.assertEqual(0, len(c))
        log_called_args = LOG.warning.call_args_list
        self.assertEqual(
            'Only 1 fetched meters contain "volume" field instead of 2.',
            log_called_args[0][0][0] % log_called_args[0][0][1])

    def test_arithmetic_expr_meter(self):
        cfg = yaml.dump(
            {'metric': [dict(name='compute.node.cpu.percent',
                        event_type="compute.metrics.update",
                        type='gauge',
                        unit="percent",
                        volume="$.payload.metrics["
                               "?(@.name='cpu.percent')].value"
                               " * 100",
                        resource_id="$.payload.host + '_'"
                                    " + $.payload.nodename")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(METRICS_UPDATE))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        self.assertEqual('compute.node.cpu.percent', s1['name'])
        self.assertEqual(2.7501485834103514, s1['volume'])
        self.assertEqual("tianst_tianst.sh.intel.com",
                         s1['resource_id'])

    def test_string_expr_meter(self):
        cfg = yaml.dump(
            {'metric': [dict(name='compute.node.cpu.frequency',
                        event_type="compute.metrics.update",
                        type='gauge',
                        unit="ns",
                        volume="$.payload.metrics[?(@.name='cpu.frequency')]"
                               ".value",
                        resource_id="$.payload.host + '_'"
                                    " + $.payload.nodename")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(METRICS_UPDATE))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        self.assertEqual('compute.node.cpu.frequency', s1['name'])
        self.assertEqual(1600, s1['volume'])
        self.assertEqual("tianst_tianst.sh.intel.com",
                         s1['resource_id'])

    def test_prefix_expr_meter(self):
        cfg = yaml.dump(
            {'metric': [dict(name='compute.node.cpu.frequency',
                        event_type="compute.metrics.update",
                        type='gauge',
                        unit="ns",
                        volume="$.payload.metrics[?(@.name='cpu.frequency')]"
                               ".value",
                        resource_id="'prefix-' + $.payload.nodename")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(METRICS_UPDATE))
        self.assertEqual(1, len(c))
        s1 = c[0].as_dict()
        self.assertEqual('compute.node.cpu.frequency', s1['name'])
        self.assertEqual(1600, s1['volume'])
        self.assertEqual("prefix-tianst.sh.intel.com",
                         s1['resource_id'])

    def test_duplicate_meter(self):
        cfg = yaml.dump(
            {'metric': [dict(name="test1",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id"),
                        dict(name="test1",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        self._load_meter_def_file(cfg)
        c = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(1, len(c))

    def test_multi_files_multi_meters(self):
        cfg1 = yaml.dump(
            {'metric': [dict(name="test1",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        cfg2 = yaml.dump(
            {'metric': [dict(name="test2",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        self._load_meter_def_file([cfg1, cfg2])
        data = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(2, len(data))
        expected_names = ['test1', 'test2']
        for s in data:
            self.assertIn(s.as_dict()['name'], expected_names)

    def test_multi_files_duplicate_meter(self):
        cfg1 = yaml.dump(
            {'metric': [dict(name="test",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        cfg2 = yaml.dump(
            {'metric': [dict(name="test",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        self._load_meter_def_file([cfg1, cfg2])
        data = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(1, len(data))
        self.assertEqual(data[0].as_dict()['name'], 'test')

    def test_multi_files_empty_payload(self):
        event = copy.deepcopy(MIDDLEWARE_EVENT)
        del event['payload']['measurements']
        cfg1 = yaml.dump(
            {'metric': [dict(name="$.payload.measurements.[*].metric.[*].name",
                             event_type="objectstore.http.request",
                             type="delta",
                             unit="$.payload.measurements.[*].metric.[*].unit",
                             volume="$.payload.measurements.[*].result",
                             resource_id="$.payload.target_id",
                             project_id="$.payload.initiator.project_id",
                             lookup="name")]})
        cfg2 = yaml.dump(
            {'metric': [dict(name="$.payload.measurements.[*].metric.[*].name",
                             event_type="objectstore.http.request",
                             type="delta",
                             unit="$.payload.measurements.[*].metric.[*].unit",
                             volume="$.payload.measurements.[*].result",
                             resource_id="$.payload.target_id",
                             project_id="$.payload.initiator.project_id",
                             lookup="name")]})
        self._load_meter_def_file([cfg1, cfg2])
        data = list(self.handler.build_sample(event))
        self.assertEqual(0, len(data))

    def test_multi_files_unmatched_meter(self):
        cfg1 = yaml.dump(
            {'metric': [dict(name="test1",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        cfg2 = yaml.dump(
            {'metric': [dict(name="test2",
                        event_type="test.update",
                        type="delta",
                        unit="B",
                        volume="$.payload.volume",
                        resource_id="$.payload.resource_id",
                        project_id="$.payload.project_id")]})
        self._load_meter_def_file([cfg1, cfg2])
        data = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(1, len(data))
        self.assertEqual(data[0].as_dict()['name'], 'test1')

    @mock.patch('ceilometer.meter.notifications.LOG')
    def test_multi_files_bad_meter(self, LOG):
        cfg1 = yaml.dump(
            {'metric': [dict(name="test1",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id"),
                        dict(name="bad_test",
                             type="bad_type",
                             event_type="bar.create",
                             unit="foo", volume="bar",
                             resource_id="bea70e51c7340cb9d555b15cbfcaec23")]})
        cfg2 = yaml.dump(
            {'metric': [dict(name="test2",
                             event_type="test.create",
                             type="delta",
                             unit="B",
                             volume="$.payload.volume",
                             resource_id="$.payload.resource_id",
                             project_id="$.payload.project_id")]})
        self._load_meter_def_file([cfg1, cfg2])
        data = list(self.handler.build_sample(NOTIFICATION))
        self.assertEqual(2, len(data))
        expected_names = ['test1', 'test2']
        for s in data:
            self.assertIn(s.as_dict()['name'], expected_names)
        args, kwargs = LOG.error.call_args_list[0]
        self.assertEqual("Error loading meter definition: %s", args[0])
        self.assertTrue(
            str(args[1]).endswith("Invalid type bad_type specified"))
